/*
 * burtBackup.c
 *
 *    This file implements the backup command.  This is the nuts and bolts
 *    of Burt's backup services.
 *
 * Copyright (C) 1995 - 1998 Eric Melski and Dean Jansa
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice and this permission notice appear in all copies,
 * and that all support documentation is included with each copy of the software.
 * A reasonable copying fee may be charged for any distribution of this 
 * software, but no fee may be charged for the software itself. You may 
 * distribute this software in aggregate with other programs as part of a 
 * larger software distribution provided that this software is not advertised 
 * as a product of your own.  Eric Melski and Dean Jansa make no representations
 * about the suitability of this software for any purpose.  It is provided 
 * "as is" without express or implied warranty.
 *
 * ERIC MELSKI AND DEAN JANSA DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL ERIC MELSKI OR DEAN JANSA BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 */


#ifndef lint
static char rcsid[] = "$Id: burtBackup.c,v 1.28 1998/10/14 17:34:34 jmelski Exp $";
#endif

#include <burtInt.h>
#include <signal.h>

typedef struct Burt_BackupMetaControlBlock {
  Tcl_Interp *interp;
  Tcl_Channel outputChannel;     /* Channel to write output.    */
  Tcl_Channel logChannel;        /* Channel to write log data.  */
} Burt_BackupMetaControlBlock;

typedef struct Burt_BackupControlBlock {
  Burt_Packet packet;            /* Packet for this backup. */
  Tcl_TimerToken sentryTimer;    /* Token to the sentry timer. */
  unsigned int sentry;           /* Holds number of bytes read from
				  * inputChannel at last sentryTimer */
  Tcl_Channel inputChannel;      /* Channel to read data from. */
  Tcl_Channel stderrChannel;     /* Channel to read stderr.    */
  char *backupMonitor;           /* Name of user def'd proc to send stderr 
                                    * to check for errors, or just dialog. */
} Burt_BackupControlBlock;

/*
 * This structure describes per-instance state of a pipe based channel.
 */

typedef struct PipeState {
  Tcl_Channel channel;  /* Channel associated with this file. */
  TclFile inFile;       /* Output from pipe. */
  TclFile outFile;	/* Input to pipe. */
  TclFile errorFile;	/* Error output from pipe. */
  int numPids;	        /* How many processes are attached to this pipe? */
  int *pidPtr;	        /* The process IDs themselves. Allocated by
		         * the creator of the pipe. */
  int isNonBlocking;	/* Nonzero when the pipe is in nonblocking mode.
                         * Used to decide whether to wait for the children
                         * at close time. */
} PipeState;

static int StartBackupChannel(void);
static void CloseBackupChannel(Burt_BackupControlBlock *backupCB);
static void AbortBackup(void);
static int GetUserBackupChannels(Tcl_Interp *interp, 
				 Burt_BackupControlBlock *backupCB);
static void ProcessBackupInputChannel(ClientData clientData, int mask);
static void ProcessBackupStderrChannel(ClientData clientData, int mask);
static void FlushBackupStderrChannel(Burt_BackupControlBlock *backupCB);
static void ProcessBackupSentry(ClientData clientData);
static int BurtBackupIsComplete(void);
static int LoadNextBackupTape(Tcl_Channel mtChannel, Tcl_Channel logChannel);
static void ProcessLoadBackupTapeTimer(ClientData clientData);
static void SetBurtTapeLabel(Burt_TapeLabel *tapeLabel);

static Tcl_HashTable backupCBTable;
static Burt_BackupMetaControlBlock backupMCB;
static char *savedBackupStatus;


/*
 *----------------------------------------------------------------------
 *
 * Burt_BackupCmd --
 *
 *	This procedure implements the Tcl command backup.
 *
 * Results:
 *      Returns a standard Tcl completion code, and leaves an error
 *	message in interp->result if an error occurs.
 *
 * Side effects:
 *	
 *
 *----------------------------------------------------------------------
 */

int 
Burt_BackupCmd(ClientData cdata, Tcl_Interp *interp, int argc, char **argv)
{
  int length;
  int numChannels;
  Tcl_Channel outputChannel;
  Tcl_Channel logChannel;
  Burt_TapeLabel tapeLabel;
  int mode;
  int i;
  int nbytes;
  Burt_Packet packet;
  char logBuffer[1024];

  if ((argc < 2) || (argc > 6)) 
  {
    Tcl_SetResult(interp, "wrong # args: should be \"backup option ?arg?\"", 
		  TCL_STATIC);
    return TCL_ERROR;
  }
 
  length = strlen(argv[1]);
  if ((strncmp(argv[1], "start", length) == 0) || 
      (strncmp(argv[1], "append", length) == 0))
  {
    if (argc != 6) 
    {
      Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0], " ",
		       argv[1], " channels outputchannel logchannel",
		       " tapelabel\"", (char *) NULL);
      return TCL_ERROR;
    }

    if (!BurtBackupIsComplete())
    {
      Tcl_SetResult(interp, "BURT: backup already running", TCL_STATIC);
      return(TCL_ERROR);
    }

    /* Initialize the backupCBTable */
    Tcl_InitHashTable(&backupCBTable, TCL_STRING_KEYS);

    if (Tcl_GetInt(interp, argv[2], &numChannels) != TCL_OK) 
      return(TCL_ERROR);
	
    outputChannel = Tcl_GetChannel(interp, argv[3], &mode);
    if (outputChannel == NULL)
      /* Tcl_GetChannel fills in the error message for us. */
      return(TCL_ERROR);

    if (((mode & TCL_WRITABLE) == 0) || ((mode & TCL_READABLE) == 0))
    {
      Tcl_SetResult(interp, "BURT: Output channel wasn't opened for "
		    "reading/writing", TCL_STATIC);
      return(TCL_ERROR);
    }     

    logChannel = Tcl_GetChannel(interp, argv[4], &mode);
    if (logChannel == NULL)
      /* Tcl_GetChannel fills in the error message for us. */
      return(TCL_ERROR);

    if ((mode & TCL_WRITABLE) == 0) 
    {
      Tcl_SetResult(interp, "BURT: Log channel wasn't opened for writing", 
		    TCL_STATIC);
      return(TCL_ERROR);
    }

    Tcl_SetVar(interp, "burt_complete", "0", TCL_GLOBAL_ONLY);
    BurtReadtapeInit();
    BurtStatusReset();
    BurtStatusSetRunState(BACKUP | RUN);
    BurtStatusSetTapeLabel(argv[5]);
    BurtStatusSetScheduled(BurtTableLength(&scheduleTable));
    BurtStatusSetChannels(numChannels);
    BurtStatusSetFinished(0);
    BurtStatusSetStarted(0);
    BurtStatusSetTapes(1);
    BurtStatusSetBytes(0);
    BurtStatusSetStartTime(time((time_t *) NULL));

    if (BurtBackupIsComplete()) 
    {
      Tcl_ResetResult(interp);
      Tcl_AppendResult(interp, "BURT: Nothing scheduled, aborting backup [",
		       BurtTime(), "]", (char *)NULL);
      BurtStatusReset();
      return(TCL_ERROR);
    }

    if (BurtMtRewind(outputChannel) != MT_OK)
    {
      Tcl_SetResult(interp, "BURT: Unable to rewind tape, aborting", TCL_STATIC);
      BurtStatusReset();
      return(TCL_ERROR);
    }

    /* Only write a new tape label if we're starting a new backup */
    if (strncmp(argv[1], "start", length) == 0)
    {
      /* Write the new tape label */
      SetBurtTapeLabel(&tapeLabel);
      nbytes = BurtMtWriteLabel(outputChannel, &tapeLabel);
      if (nbytes == MT_ERROR)
      {
	Tcl_SetResult(interp, "BURT: Could not write tape label", TCL_STATIC);
	BurtStatusReset();
	return(TCL_ERROR); 
      }
      /* BurtStatusSetBytes(nbytes); */
    }
    else if (strncmp(argv[1], "append", length) == 0)
    {
      /* If we're appending, verify that this tape can be appended to */
      sprintf(logBuffer, "BURT: Verifying BURT version on tape... [%s]\n",
	      BurtTime());
      Tcl_Write(logChannel, logBuffer, -1);
      Tcl_Flush(logChannel);
      if (BurtMtReadLabel(outputChannel, &tapeLabel) != MT_OK)
      {
	Tcl_SetResult(interp, "BURT: Unable to read label, aborting", TCL_STATIC);
	BurtStatusReset();
	return(TCL_ERROR);
      }
      
      if (tapeLabel.magic < 1500)  /* v1.5 == first to support backup append */ 
      {
	Tcl_SetResult(interp, "BURT: Tape was written with a BURT version which "
		      "did not support backup append.  Aborting.", TCL_STATIC);
	BurtStatusReset();
	return(TCL_ERROR);
      }
      
      if (BurtMtRewind(outputChannel) != MT_OK)
      {
	Tcl_SetResult(interp, "BURT: Unable to rewind tape, aborting", 
		      TCL_STATIC);
	BurtStatusReset();
	return(TCL_ERROR);
      }

      sprintf(logBuffer, "BURT: Seeking end of media... [%s]\n", BurtTime());
      Tcl_Write(logChannel, logBuffer, -1);
      Tcl_Flush(logChannel);
      if (BurtMtEOM(outputChannel) != MT_OK)
      {
	Tcl_SetResult(interp, "BURT: Unable to find EOM, aborting", TCL_STATIC);
	BurtStatusReset();
	return(TCL_ERROR);
      }
      
      sprintf(logBuffer, "BURT: Located end of media [%s]\n", BurtTime());
      Tcl_Write(logChannel, logBuffer, -1);
      sprintf(logBuffer, "BURT: Seeking END_OF_BACKUPSET packet... [%s]\n", 
	      BurtTime());
      Tcl_Write(logChannel, logBuffer, -1);
      Tcl_Flush(logChannel);
      if (BurtMtBSF(outputChannel, 2) != MT_OK)
      {
	Tcl_SetResult(interp, "BURT: Unable to backspace filemarks, aborting", 
		      TCL_STATIC);
	BurtStatusReset();
	return(TCL_ERROR);
      }
      
      nbytes = BurtMtRead(outputChannel, &packet);
      if (nbytes != sizeof(Burt_Packet))
      {
	Tcl_SetResult(interp, "BURT: Unable to read end of backupset packet", 
		      TCL_STATIC);
	BurtStatusReset();
	return(TCL_ERROR);
      }
      
      if (packet.blockSize != END_OF_BACKUPSET)
      {
	Tcl_SetResult(interp, "BURT: Unable to find end of backupset packet",
		      TCL_STATIC);
	BurtStatusReset();
	return(TCL_ERROR);
      }

      if (BurtMtBSF(outputChannel, 1) != MT_OK)
      {
	Tcl_SetResult(interp, "BURT: Unable to backspace filemarks, aborting", 
		      TCL_STATIC);
	BurtStatusReset();
	return(TCL_ERROR);
      }
      sprintf(logBuffer, "BURT: Located END_OF_BACKUPSET packet [%s]\n", 
	      BurtTime());
      Tcl_Write(logChannel, logBuffer, -1);
      Tcl_Flush(logChannel);
    }
    
    /* EVERYBODY (ie append and start) fires up a bunch of backup channels */
    backupMCB.interp = interp;
    backupMCB.outputChannel = outputChannel;
    backupMCB.logChannel = logChannel;
    for (i = 0; i < numChannels; i++) 
    {
      if (StartBackupChannel() == TCL_ERROR) {
	/* 
	 * StartBackupChannel fills in the error condition.
	 */
	BurtStatusReset();
	return(TCL_ERROR);
      }
    }
    
    sprintf(interp->result, "BURT: Started %d channels [%s]", numChannels, 
	    BurtTime());
    return TCL_OK;
  }
  else if (strncmp(argv[1], "stop", length) == 0) 
  {
    /* 
     * Set runState to STOP, which allows remaining (currently running)
     * channels to finish up before setting burt_complete. 
     */
    BurtStatusSetRunState(BACKUP | STOP);
    Tcl_SetResult(interp, "stopped, draining remaining channels", TCL_STATIC);
    return(TCL_OK);
  } 
  else if (strncmp(argv[1], "abort", length) == 0) 
  {
    /* 
     * Ouch..... 
     */
    if (BurtBackupIsComplete())
    {
      Tcl_SetResult(interp, "BURT: No backup running, cannot abort.", TCL_STATIC);
      return(TCL_ERROR);
    }

    sprintf(logBuffer, "BURT: Received backup abort command, aborting [%s]\n",
	    BurtTime());
    Tcl_Write(backupMCB.logChannel, logBuffer, -1);
    Tcl_Flush(backupMCB.logChannel);
    
    AbortBackup();
    return(TCL_OK);
  } 
  else 
  {
    Tcl_AppendResult(interp, "bad option \"", argv[1], 
		     "\": should be start, stop, abort or append", 
		     (char *) NULL);
    return(TCL_ERROR);    
  }
}


/*
 *----------------------------------------------------------------------
 *
 * StartBackupChannel --
 *
 *	This procedure sets up a new channel handler for a scheduled Item.
 *      After getting the next item from the schedule table it arranges 
 *      for the channels to be processed whenever the given channel has 
 *      data waiting to be read.  During this time a Burt_BackupControlBlock 
 *      is built, which is passed along to the channel handlers.
 *
 * Results:
 *	Returns TCL_OK on succes, or TCL_ERROR on failure.  Fills in
 *      interp->result with info upon error.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */

static int 
StartBackupChannel(void)
{
  Burt_ScheduleItem *scheduleItemPtr;
  Burt_Packet packet;
  Burt_BackupControlBlock *backupCB;
  Tcl_HashEntry *entryPtr;
  int result;
  char buffer[80];
  char key[BURT_MAX_HOSTLEN + BURT_MAX_ATOMLEN + 
	  BURT_MAX_LEVELLEN + BURT_MAX_TYPELEN + 1];
  int new;

  scheduleItemPtr = BurtScheduleGetNext(backupMCB.interp, backupMCB.logChannel,
					&scheduleTable);
  if (scheduleItemPtr != NULL) 
  {
    backupCB = (Burt_BackupControlBlock *)malloc(sizeof(Burt_BackupControlBlock));
    if (backupCB == NULL) 
    {
      Tcl_SetResult(backupMCB.interp, "Out of Memory in StartBackupChannel", 
		    TCL_STATIC);
      return TCL_ERROR;
    }
	
    backupCB->packet = packet;
    backupCB->packet.magic = BURT_MAGIC_NUMBER;
    strncpy(backupCB->packet.host, scheduleItemPtr->host, BURT_MAX_HOSTLEN);
    strncpy(backupCB->packet.atom, scheduleItemPtr->atom, BURT_MAX_ATOMLEN);
    strncpy(backupCB->packet.type, scheduleItemPtr->type, BURT_MAX_TYPELEN);
    strncpy(backupCB->packet.level, scheduleItemPtr->level, BURT_MAX_LEVELLEN);
    backupCB->packet.blockNum = 0;
    backupCB->packet.blockSize = 0;
    backupCB->packet.totalSize = 0;
    backupCB->packet.checksum = 0;
    free((void *) scheduleItemPtr);

    if (GetUserBackupChannels(backupMCB.interp, backupCB) == TCL_ERROR) 
      return(TCL_ERROR);

    result = Tcl_SetChannelOption(backupMCB.interp, backupCB->inputChannel, 
				  "-blocking", "false");
    if (result == TCL_ERROR) 
      return(TCL_ERROR);

    result = Tcl_SetChannelOption(backupMCB.interp, backupCB->inputChannel, 
				  "-translation", "binary");
    if (result == TCL_ERROR)
      return(TCL_ERROR);

    result = Tcl_SetChannelOption(backupMCB.interp, backupCB->stderrChannel, 
				  "-blocking", "false");
    if (result == TCL_ERROR)
      return(TCL_ERROR);

    /* 
     * Set up all of the event handlers, and drop a bread crumb before 
     * the start of this backup.  Add the backupCB to the backupCBTable.
     */

    strncpy(key, backupCB->packet.host, BURT_MAX_HOSTLEN);
    strcat(key, backupCB->packet.atom);
    strcat(key, backupCB->packet.level);
    strcat(key, backupCB->packet.type);

    entryPtr = Tcl_CreateHashEntry(&backupCBTable, key, &new);

    if (new)
    {
      Tcl_SetHashValue(entryPtr, backupCB);
    }
    else
    {
      return (TCL_ERROR);
    }

    Tcl_CreateChannelHandler(backupCB->inputChannel, TCL_READABLE, 
			     ProcessBackupInputChannel, backupCB);
    Tcl_CreateChannelHandler(backupCB->stderrChannel, TCL_READABLE, 
			     ProcessBackupStderrChannel, backupCB); 
    backupCB->sentry = -1;
    backupCB->sentryTimer = Tcl_CreateTimerHandler(BURT_SENTRY_TIMEOUT, 
						   ProcessBackupSentry, 
						   (ClientData) backupCB);

    BurtStatusSetStarted(BurtStatusGetStarted() + 1);

    sprintf(buffer, "BURT: Started %s %s %s %s [%s]\n",
	    backupCB->packet.host, backupCB->packet.atom, 
	    backupCB->packet.level, backupCB->packet.type, BurtTime());
    Tcl_Write(backupMCB.logChannel, buffer, strlen(buffer));
    Tcl_Flush(backupMCB.logChannel);
  } 
  return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * CloseBackupChannel --
 *
 *	This procedure closes a backup channel, it deletes all of the
 *      event handlers associated with the backup, frees the backupCB,
 *      and attempts to start a new backup channel to replace the one
 *      which was just closed.  It honors the run state, so if burt is
 *      not in the "RUN" run state it will only do the close, and skip
 *      the starting of a  new channel.  If there are no more backups
 *      to start up, it will write out the final END_OF_BACKUPSET to the
 *      tape to make the end of the run.
 *
 * Results:
 *
 * Side effects:
 *     New backup channel may be started.
 *----------------------------------------------------------------------
 */

static void 
CloseBackupChannel(Burt_BackupControlBlock *backupCB)
{
  Tcl_ChannelType *chanTypePtr;	/* The type of that channel. */
  PipeState *pipePtr;     /* The state of the pipe handed back user. */
  register int i;
  char *cleanupProc;
  char logBuffer[1024];
  char key[BURT_MAX_HOSTLEN + BURT_MAX_ATOMLEN + 
	  BURT_MAX_LEVELLEN + BURT_MAX_TYPELEN + 1];
  Tcl_HashEntry *entryPtr;

  BurtStatusSetFinished(BurtStatusGetFinished() + 1);

  /*
   * If the user gave us pipes, kill the associated processes
   * before we close up the pipes, this must be done to prevent
   * the processes (namely rsh) from hanging.
   */
  
  chanTypePtr = Tcl_GetChannelType(backupCB->inputChannel);
  if (strcmp(chanTypePtr->typeName, "pipe") == 0) 
  {
    pipePtr = (PipeState *) Tcl_GetChannelInstanceData(backupCB->inputChannel);
    for (i = 0; i < pipePtr->numPids; i++) 
    {
      kill((int)(pipePtr->pidPtr[i]), SIGKILL);
    }
  }

  /* 
   * Delete all the event handlers. Close up input and stderr channels.
   */

  Tcl_DeleteChannelHandler(backupCB->inputChannel, ProcessBackupInputChannel,
			   backupCB);
  Tcl_UnregisterChannel(backupMCB.interp, backupCB->inputChannel);
  Tcl_DeleteChannelHandler(backupCB->stderrChannel, 
			   ProcessBackupStderrChannel, backupCB);
  Tcl_UnregisterChannel(backupMCB.interp, backupCB->stderrChannel);
  Tcl_DeleteTimerHandler(backupCB->sentryTimer);
  
  /*
   * Call the cleanup procedure for this type, if one is registered
   */
  cleanupProc = Tcl_GetVar2(backupMCB.interp, "burt_cleanup_proc",
			    backupCB->packet.type, TCL_GLOBAL_ONLY);
  if (cleanupProc != NULL)
  {
    if (Tcl_VarEval(backupMCB.interp, cleanupProc, " ", 
		    backupCB->packet.host, " ", backupCB->packet.atom, 
		    " ", backupCB->packet.level, NULL) == TCL_ERROR)
    {
      sprintf(logBuffer, "BURT: Error detected in user defined cleanup "
	      "procedure %s: ", cleanupProc);
      Tcl_Write(backupMCB.logChannel, logBuffer, -1);
      Tcl_Write(backupMCB.logChannel, backupMCB.interp->result, -1);
      sprintf(logBuffer, " [%s]\n", BurtTime());
      Tcl_Write(backupMCB.logChannel, logBuffer, -1);
      Tcl_Flush(backupMCB.logChannel);
    }
  }

  /*
   * Remove the backupCB entry from the backupCBTable 
   */
  
  strncpy(key, backupCB->packet.host, BURT_MAX_HOSTLEN);
  strcat(key, backupCB->packet.atom);
  strcat(key, backupCB->packet.level);
  strcat(key, backupCB->packet.type);

  entryPtr = Tcl_FindHashEntry(&backupCBTable, key);
  if (entryPtr != NULL)
    Tcl_DeleteHashEntry(entryPtr);

  /* 
   * If run state is STOP or ABORT, we don't want to start any more channels.
   */
    
  if ((BurtStatusGetRunState() & STATEMASK) != STOP && 
      (BurtStatusGetRunState() & STATEMASK) != ABORT)
  {
    if (StartBackupChannel() == TCL_ERROR) 
    {
      Tcl_BackgroundError(backupMCB.interp);
    }
  }

  /*
   * Check to see if all of the work is done, and set things to
   * reflect what we find.
   */
	
  if (BurtBackupIsComplete()) 
  {
    backupCB->packet.blockSize = END_OF_BACKUPSET;
    backupCB->packet.blockNum++;
    backupCB->packet.checksum = 0;
    /* AGG, must check for EOT here... */
    /* Write a filemark so we can seek to EOS (End Of Set) for appending */
    BurtMtWriteFilemark(backupMCB.outputChannel);
    BurtMtWrite(backupMCB.outputChannel, &(backupCB->packet));
    BurtStatusSetRunState(BACKUP | COMPLETE);
    BurtStatusSetCompleteTime(time((time_t *) NULL));
    Tcl_DeleteHashTable(&backupCBTable);
    Tcl_SetVar(backupMCB.interp, "burt_complete", "1", TCL_GLOBAL_ONLY);
  } 
  free((void *) backupCB);
}


/*
 *----------------------------------------------------------------------
 *
 * AbortBackup --
 *
 *	This procedure aborts all running backup channels by repeatedly
 *      calling CloseBackupChannel, stamps the log, sets the run state to ABORT,
 *      and sets burt_complete
 *
 * Results:
 *
 * Side effects:
 *      All backup channels are closed; burt_complete is set to 2, 
 *      backupCBTable is deleted, run state is set to ABORT.
 *----------------------------------------------------------------------
 */

static void 
AbortBackup(void)
{
  Burt_BackupControlBlock *backupCB;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch searchPtr;
  char logBuffer[1024];

  BurtStatusSetRunState(BACKUP | ABORT);

  entryPtr = Tcl_FirstHashEntry(&backupCBTable, &searchPtr);

  while (entryPtr != NULL)
  {
    backupCB = (Burt_BackupControlBlock *)Tcl_GetHashValue(entryPtr);
    CloseBackupChannel(backupCB);
    entryPtr = Tcl_NextHashEntry(&searchPtr);
  }

  sprintf(logBuffer, "BURT: Backups aborted [%s]\n", BurtTime());
  Tcl_Write(backupMCB.logChannel, logBuffer, -1);
  Tcl_Flush(backupMCB.logChannel);
  BurtStatusSetCompleteTime(time((time_t *) NULL));
  Tcl_DeleteHashTable(&backupCBTable);
  Tcl_SetVar(backupMCB.interp, "burt_complete", "2", TCL_GLOBAL_ONLY);
}


/*
 *----------------------------------------------------------------------
 *
 * GetUserBackupChannels --
 *
 *	This procedure fills in the inputChannel, stderrChannel, and 
 *      backupMonitor fields of the backup control block.  It gets the 
 *      inputChannel and stderrChannel  by calling a user defined 
 *      TCL procedure (looked up in a Tcl array based on the "type" field 
 *      of the control block.) which must return two readable channels.  
 *      The first channel is considered stdout or the inputChannel (This is 
 *      where the data which ends up on tape comes from) and the second 
 *      channel is the stderrChannel.  The backupMonitor is looked up
 *      in a Tcl array based on the "type" field of the control block.
 *
 * Results:
 *	Returns TCL_OK if both channels were gathered, TCL_ERROR otherwise.
 *      If TCL_ERROR is returned, the interp->result field is filled
 *      in with an error message.
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

static int
GetUserBackupChannels(Tcl_Interp *interp, Burt_BackupControlBlock *backupCB)
{
  char *userProc;
  int mode;
  int argc;
  char **argv;
  
  userProc = Tcl_GetVar2(interp, "burt_backup_proc", backupCB->packet.type, 
			 TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
  if (userProc == NULL) 
  {
    /*
     * Tcl_GetVar2 already filled in the error, couldn't find
     * the user defined procedure. 
     */
    return(TCL_ERROR);
  }

  backupCB->backupMonitor = Tcl_GetVar2(interp, "burt_backupmonitor_proc",
					backupCB->packet.type,
					TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);

  if (backupCB->backupMonitor == NULL) 
  {
    /*
     * Tcl_GetVar2 already filled in the error, couldn't find
     * the user defined procedure. 
     */
    return(TCL_ERROR);
  }

  if (Tcl_VarEval(interp, userProc, " ", backupCB->packet.host, " ",
		  backupCB->packet.atom, " ", backupCB->packet.level, 
		  NULL) == TCL_ERROR) 
  {
    /* 
     * Tcl_VarEval found an error in user proc and filled in the error.
     */
    return(TCL_ERROR);
  }

  /*
   * User procedure returns a list of channels, process them both.
   */

  if (Tcl_SplitList(interp, interp->result, &argc, &argv) == TCL_ERROR) 
    return(TCL_ERROR);
  
  if (argc != 2) 
  {
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp, "User defined procedure ", userProc,
		     " must return a Tcl list with 2 elements.", (char *) NULL);
    return(TCL_ERROR);
  }
	
  backupCB->inputChannel = Tcl_GetChannel(interp, argv[0], &mode);
  if (backupCB->inputChannel == NULL) 
  {
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp, "can not find channel named \"", argv[0], 
		     "\" in user defined procedure", userProc, (char *) NULL);
    return(TCL_ERROR);
    
  }

  if ((mode & TCL_READABLE) == 0) 
  {
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp, "channel ", argv[0], " wasn't opened for reading", 
		     " in user defined procedure ", userProc, (char *) NULL);
    Tcl_UnregisterChannel(interp, backupCB->inputChannel);
    return(TCL_ERROR);
  }     
    
  backupCB->stderrChannel = Tcl_GetChannel(interp, argv[1], &mode);
  if (backupCB->stderrChannel == NULL) 
  {
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp, "can not find channel named \"", argv[1], 
		     "\" in user defined procedure", userProc, (char *) NULL);
    return(TCL_ERROR);
  }
  
  if ((mode & TCL_READABLE) == 0) 
  {
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp, "channel ", argv[1], " wasn't opened for reading", 
		     " in user defined procedure ", userProc, (char *) NULL);
    Tcl_UnregisterChannel(interp, backupCB->stderrChannel);
    return(TCL_ERROR);
  }   
  return(TCL_OK);
}


/*
 *----------------------------------------------------------------------
 *
 * ProcessBackupInputChannel --
 *
 *	This procedure reads a block of data from the inputChannel of
 * the backup and arranges to get the data out onto the outputChannel.
 * It updates the packet header before writing the header, and data
 * block.  It is called from a Channel Handler, so there should be data 
 * waiting when it gets here.
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

static void  
ProcessBackupInputChannel(ClientData clientData, int mask)
{
  Burt_BackupControlBlock *backupCB;
  Burt_Packet EOTPacket;
  static char buffer[1024];
  int nbytes;    /* Number of bytes read from inputChannel. */
  int mtbytes;   /* Number of bytes written to tape. */

  backupCB = (Burt_BackupControlBlock *) clientData;
  
  nbytes = Tcl_Read(backupCB->inputChannel,
		    &(backupCB->packet.buffer[backupCB->packet.blockSize]),
		    BURT_MAX_BLOCKSIZE - 580 - backupCB->packet.blockSize);
  
  if (nbytes > 0)
  {
    backupCB->packet.blockSize += nbytes;
    backupCB->packet.totalSize += nbytes;
  }
  else if (nbytes < 0) 
  {
    Tcl_BackgroundError(backupMCB.interp);
    /*
     * Kill this backup?  Yes.  Notify user, call it ?pipe error?
     */
    backupCB->packet.blockNum++;
    backupCB->packet.blockSize = ABORTED_BACKUP;
    backupCB->packet.checksum = 0; 
    
    mtbytes = BurtMtWrite(backupMCB.outputChannel, &(backupCB->packet));
    BurtStatusSetBytes(BurtStatusGetBytes() + mtbytes);
    
    FlushBackupStderrChannel(backupCB);
    sprintf(buffer, "BURT: Aborted %s %s %s %s, pipe error [%s]\n",
	    backupCB->packet.host, backupCB->packet.atom,
	    backupCB->packet.level, backupCB->packet.type, BurtTime());
    Tcl_Write(backupMCB.logChannel, buffer, strlen(buffer));
    Tcl_Flush(backupMCB.logChannel);
    
    CloseBackupChannel(backupCB);
    return;
  }

  if ((backupCB->packet.blockSize == BURT_MAX_BLOCKSIZE - 580 ||
      Tcl_Eof(backupCB->inputChannel)) && backupCB->packet.blockSize > 0)
  {
    backupCB->packet.blockNum++;
    backupCB->packet.checksum = BurtChecksum(backupCB->packet.buffer, 
					     backupCB->packet.blockSize);
    
    if (backupCB->packet.blockNum == 1) 
    {
      BurtMtWriteFilemark(backupMCB.outputChannel);
    }

    mtbytes = BurtMtWrite(backupMCB.outputChannel, &(backupCB->packet));
		    
    if (mtbytes == MT_EOT) 
    {
      do 
      {
	/* Write a LOGICAL_EOT packet header, so that readtape and 
	 * recover know this is the end of the tape.  Theoretically 
	 * they will "just know" where the end is anyway, but this 
	 * seals the deal.
	 */
	EOTPacket.blockNum = -1;
	EOTPacket.blockSize = LOGICAL_EOT;
	EOTPacket.totalSize = 0;
	EOTPacket.host[0] = '\0';
	EOTPacket.atom[0] = '\0';
	EOTPacket.type[0] = '\0';
	EOTPacket.level[0] = '\0';
	EOTPacket.checksum = 0;
	
	/* Don't have to make sure it wrote, because if it didn't, 
	 * then readtape/recover will determine EOT by the fact 
	 * that we are at physical EOT.
	 */
	BurtMtWrite(backupMCB.outputChannel, &EOTPacket);
	
	if (LoadNextBackupTape(backupMCB.outputChannel, 
			       backupMCB.logChannel) == TCL_ERROR) 
	{
	  sprintf(buffer, "BURT: I/O error, couldn't load next tape [%s]\n",
		  BurtTime());

	  Tcl_Write(backupMCB.logChannel, buffer, strlen(buffer));
	  Tcl_Flush(backupMCB.logChannel);
	  AbortBackup();
	  /* exit(EXIT_FAILURE); */
	  break;
	}
		
	/* Re-write the packet header and packet */
	mtbytes = BurtMtWrite(backupMCB.outputChannel, &(backupCB->packet));

	if (mtbytes == MT_ERROR) 
	{
	  sprintf(buffer, "BURT: Tape error, entire backup aborted [%s]\n",
		  BurtTime());
	  Tcl_Write(backupMCB.logChannel, buffer, strlen(buffer));
	  Tcl_Flush(backupMCB.logChannel);
	  AbortBackup();
	  /* exit(EXIT_FAILURE); */
	  break;
	}			      
      } 
      while (mtbytes == MT_EOT);
      BurtStatusSetBytes(BurtStatusGetBytes() + mtbytes);
      backupCB->packet.blockSize = 0;
    } 
    else if (mtbytes == MT_ERROR) 
    {
      sprintf(buffer, "BURT: Tape error, entire backup aborted [%s]\n", 
	      BurtTime());
      Tcl_Write(backupMCB.logChannel, buffer, strlen(buffer));
      Tcl_Flush(backupMCB.logChannel);
      AbortBackup();
      /* exit(EXIT_FAILURE); */
    } 
    else 
    {
      BurtStatusSetBytes(BurtStatusGetBytes() + mtbytes);
      /* reset the block size (ie, number of bytes collected so far) */
      backupCB->packet.blockSize = 0;
    }
  }

  if (Tcl_Eof(backupCB->inputChannel)) 
  {
    /*
     * Write out a "End of Backup Packet Header."
     */
    backupCB->packet.blockNum++;
    backupCB->packet.blockSize = END_OF_BACKUP;
    backupCB->packet.checksum = 0; 
    
    /* This needs to be here to keep the filemarkers in order
     * In the case of a single block dump, where the first header is
     * also the last, IE an error, such and network death, etc...
     */
    
    if (backupCB->packet.blockNum == 1) 
    {
      backupCB->packet.blockSize = ABORTED_BACKUP;
      backupCB->packet.checksum = 0;
      BurtMtWriteFilemark(backupMCB.outputChannel);
    }

    mtbytes = BurtMtWrite(backupMCB.outputChannel, &(backupCB->packet));
    BurtStatusSetBytes(BurtStatusGetBytes() + mtbytes);
    
    if (backupCB->packet.blockNum == 1) 
    {
      FlushBackupStderrChannel(backupCB);
      sprintf(buffer, "BURT: Aborted %s %s %s %s, 0 bytes and EOF on first "
	      "block of input [%s]\n", backupCB->packet.host, 
	      backupCB->packet.atom, backupCB->packet.level, 
	      backupCB->packet.type, BurtTime());
      Tcl_Write(backupMCB.logChannel, buffer, strlen(buffer));
      Tcl_Flush(backupMCB.logChannel);
    } 
    else 
    {
      sprintf(buffer, 
	      "BURT: Completed %s %s %s %s, %d blocks, totaling %u bytes [%s]\n",
	      backupCB->packet.host, backupCB->packet.atom, 
	      backupCB->packet.level, backupCB->packet.type,
	      backupCB->packet.blockNum, backupCB->packet.totalSize, 
	      BurtTime());
      Tcl_Write(backupMCB.logChannel, buffer, strlen(buffer));
      Tcl_Flush(backupMCB.logChannel);
    }
    CloseBackupChannel(backupCB);
  }
  return;
}


/*
 *----------------------------------------------------------------------
 *
 * ProcessBackupStderrChannel --
 *
 *	This procedure reads a stderr stream and re-writes the error to
 *      the logChannel with the dump stamp (I.E. the /@stall: identifier) 
 *      attached. It also calls the user defined procedure named in the
 *      backupMonitor field of the BackupControlBlock to check to see if
 *      the message was an error, or just plain old dialog.  If that proc
 *      returns anything but "0" the backup channels are closed down and
 *      a message regarding such action is placed in the log file.
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

static void  
ProcessBackupStderrChannel(ClientData clientData, int mask)
{
  Burt_BackupControlBlock *backupCB;
  Tcl_DString buffer;
  Tcl_DString logMessage;
  int nbytes;
  
  backupCB = (Burt_BackupControlBlock *) clientData;
  
  Tcl_DStringInit(&buffer);    
  nbytes = Tcl_Gets(backupCB->stderrChannel, &buffer);
  
  if (nbytes > 0) 
  {
    Tcl_DStringInit(&logMessage);    
    Tcl_DStringAppend(&logMessage, backupCB->packet.host, 
		      strlen(backupCB->packet.host));
    Tcl_DStringAppend(&logMessage, " ", -1);
    Tcl_DStringAppend(&logMessage, backupCB->packet.atom, 
		      strlen(backupCB->packet.atom));
    Tcl_DStringAppend(&logMessage, " ", -1);
    Tcl_DStringAppend(&logMessage, backupCB->packet.level, 
		      strlen(backupCB->packet.level));
    Tcl_DStringAppend(&logMessage, " ", -1);
    Tcl_DStringAppend(&logMessage, backupCB->packet.type, 
		      strlen(backupCB->packet.type));
    Tcl_DStringAppend(&logMessage, ": ", -1);
    Tcl_DStringAppend(&logMessage, Tcl_DStringValue(&buffer), nbytes);
    Tcl_DStringAppend(&logMessage, "\n", -1);
    Tcl_Write(backupMCB.logChannel, Tcl_DStringValue(&logMessage), 
	      Tcl_DStringLength(&logMessage));
    Tcl_Flush(backupMCB.logChannel);
    Tcl_DStringFree(&logMessage);
    
    /*
     * Call user defined proc to see if that log message should be
     * considered an error, or just plain dialog.
     */
    if (Tcl_VarEval(backupMCB.interp, backupCB->backupMonitor, " ", 
		    backupCB->packet.host, " ", backupCB->packet.atom, " ",
		    backupCB->packet.level, " ", 
		    "{", Tcl_DStringValue(&buffer), "}", NULL) == TCL_ERROR) {
      /* 
       * Tcl_VarEval found an error in user proc and filled in the error.
       */
      Tcl_BackgroundError(backupMCB.interp);
    } 
    else 
    {
      if (strcmp(backupMCB.interp->result, "0") != 0) 
      {
	backupCB->packet.blockNum++;
	backupCB->packet.blockSize = ABORTED_BACKUP;
	backupCB->packet.checksum = 0; 
	      
	nbytes = BurtMtWrite(backupMCB.outputChannel, &(backupCB->packet));
	BurtStatusSetBytes(BurtStatusGetBytes() + nbytes);
	
	FlushBackupStderrChannel(backupCB);
	Tcl_DStringInit(&logMessage); 
	Tcl_DStringAppend(&logMessage, "BURT: Aborted ", -1);
	Tcl_DStringAppend(&logMessage, backupCB->packet.host, 
			  strlen(backupCB->packet.host));
	Tcl_DStringAppend(&logMessage, " ", -1);
	Tcl_DStringAppend(&logMessage, backupCB->packet.atom, 
			  strlen(backupCB->packet.atom));
	Tcl_DStringAppend(&logMessage, " ", -1);
	Tcl_DStringAppend(&logMessage, backupCB->packet.level, 
			  strlen(backupCB->packet.level));
	Tcl_DStringAppend(&logMessage, " ", -1);
	Tcl_DStringAppend(&logMessage, backupCB->packet.type, 
			  strlen(backupCB->packet.type));
	Tcl_DStringAppend(&logMessage, 
			  ", user defined procedure ", -1);  
	Tcl_DStringAppend(&logMessage, backupCB->backupMonitor, 
			  strlen(backupCB->backupMonitor));
	Tcl_DStringAppend(&logMessage, " detected an error [", -1);
	Tcl_DStringAppend(&logMessage, BurtTime(), -1);
	Tcl_DStringAppend(&logMessage, "]\n", -1);
	Tcl_Write(backupMCB.logChannel, Tcl_DStringValue(&logMessage), 
		  Tcl_DStringLength(&logMessage));
	Tcl_DStringFree(&logMessage);
	Tcl_Flush(backupMCB.logChannel);
	
	CloseBackupChannel(backupCB);
      }
    }
  }
  
  Tcl_DStringFree(&buffer);
  return;
}


/*
 *----------------------------------------------------------------------
 *
 * FlushBackupStderrChannel --
 *
 *	This procedure flushes all available input on the stderr channel.
 *      It is meant to be used right before an atom is aborted, to ensure
 *      that all information from the stderr channel makes it into the log.
 *      Otherwise, when CloseBackupChannel is called, the errors are lost.
 *      It does no checking of the error messages, since the atom is already
 *      going to be aborted anyway.
 *
 * Results:
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

static void
FlushBackupStderrChannel(Burt_BackupControlBlock *backupCB)
{
  Tcl_DString buffer;
  Tcl_DString logMessage;
  int nbytes;
  
  while (Tcl_InputBuffered(backupCB->stderrChannel) > 0 && 
	 !Tcl_Eof(backupCB->stderrChannel)) 
  {
    Tcl_DStringInit(&buffer);    
    nbytes = Tcl_Gets(backupCB->stderrChannel, &buffer);
  
    if (nbytes > 0) 
    {
      Tcl_DStringInit(&logMessage);    
      Tcl_DStringAppend(&logMessage, backupCB->packet.host, 
			strlen(backupCB->packet.host));
      Tcl_DStringAppend(&logMessage, " ", -1);
      Tcl_DStringAppend(&logMessage, backupCB->packet.atom, 
			strlen(backupCB->packet.atom));
      Tcl_DStringAppend(&logMessage, " ", -1);
      Tcl_DStringAppend(&logMessage, backupCB->packet.level, 
			strlen(backupCB->packet.level));
      Tcl_DStringAppend(&logMessage, " ", -1);
      Tcl_DStringAppend(&logMessage, backupCB->packet.type, 
			strlen(backupCB->packet.type));
      Tcl_DStringAppend(&logMessage, ": ", -1);
      Tcl_DStringAppend(&logMessage, Tcl_DStringValue(&buffer), nbytes);
      Tcl_DStringAppend(&logMessage, "\n", -1);
      Tcl_Write(backupMCB.logChannel, Tcl_DStringValue(&logMessage), 
		Tcl_DStringLength(&logMessage));
      Tcl_Flush(backupMCB.logChannel);
      Tcl_DStringFree(&logMessage);
      Tcl_DStringFree(&buffer);
    }
    else
    {
      Tcl_DStringFree(&buffer);
      break;
    }
  }
  return;
}


/*
 *----------------------------------------------------------------------
 *
 * ProcessBackupSentry --
 *
 *	This procedure determines if a backup is stalled.  It is called
 *      every X seconds and confirms that data is moving thru the 
 *      inputChannel.  If no data movement is found after a set time the
 *      backup is killed.
 *
 * Results:
 *
 * Side effects:
 *      Normally just resets itself to be called in the future, but may
 *      kill the backup if it appears stalled.
 *----------------------------------------------------------------------
 */

static void 
ProcessBackupSentry(ClientData clientData)
{
  Burt_BackupControlBlock *backupCB;
  Tcl_DString buffer;
  int mtbytes;

  backupCB = (Burt_BackupControlBlock *) clientData;
  Tcl_DStringInit(&buffer);
  
  if ((Tcl_InputBuffered(backupCB->inputChannel) == 0) 
      && (backupCB->packet.totalSize == backupCB->sentry)) 
  {
    /* 
     * Stalled, close up the backup. 
     */
    backupCB->packet.blockNum++;
    backupCB->packet.blockSize = KILLED_BACKUP;
    backupCB->packet.checksum = 0; 
    
    mtbytes = BurtMtWrite(backupMCB.outputChannel, &(backupCB->packet));
    BurtStatusSetBytes(BurtStatusGetBytes() + mtbytes);

    Tcl_DStringAppend(&buffer, "BURT: Killed ", -1);
    Tcl_DStringAppend(&buffer, backupCB->packet.host, -1);
    Tcl_DStringAppend(&buffer, " ", -1);
    Tcl_DStringAppend(&buffer, backupCB->packet.atom, -1);
    Tcl_DStringAppend(&buffer, " ", -1);
    Tcl_DStringAppend(&buffer, backupCB->packet.level, -1);
    Tcl_DStringAppend(&buffer, " ", -1);
    Tcl_DStringAppend(&buffer, backupCB->packet.type, -1);
    Tcl_DStringAppend(&buffer, ", Sentry Timeout [", -1);
    Tcl_DStringAppend(&buffer, BurtTime(), -1);
    Tcl_DStringAppend(&buffer, "]\n", -1);
    Tcl_Write(backupMCB.logChannel, Tcl_DStringValue(&buffer), 
	      Tcl_DStringLength(&buffer));
    
    CloseBackupChannel(backupCB);
  } 
  else 
  {
    /* 
     * Reset sentryTimer 
     */
    backupCB->sentry = backupCB->packet.totalSize;
    backupCB->sentryTimer = Tcl_CreateTimerHandler(BURT_SENTRY_TIMEOUT, 
						   ProcessBackupSentry, 
						   (ClientData) backupCB);
  }

  Tcl_DStringFree(&buffer);
  return;
}


/*
 *----------------------------------------------------------------------
 *
 * BurtBackupIsComplete --
 *
 *	This procedure checks to see if there is any outstanding
 *      work to be done.
 *
 * Results:
 *      1 if there is nothing left to do, otherwise 0.
 *
 * Side effects:
 *      None.
 *----------------------------------------------------------------------
 */

static int
BurtBackupIsComplete(void)
{
  /*
   * There are three cases to check to see if Burt has completed 
   * its work:
   * 1) Run state = RUN, and burtBackupStatus.finished 
   *                             == burtBackupStatus.scheduled
   *    (I.E. the schedule Q is empty, and all the channels have finished)
   *
   * 2) Run state = STOP, and burtBackupStatus.finished 
   *                              == burtBackupStatus.started
   *    (I.E. the outstanding channels have finished up)
   * 3) Run state = COMPLETE (I.E., a backup has already run)
   */
  
  if ((((BurtStatusGetRunState() & STATEMASK) == RUN) 
       && (BurtStatusGetFinished() == BurtStatusGetScheduled()))
      || (((BurtStatusGetRunState() & STATEMASK) == STOP) 
	  && (BurtStatusGetFinished() == BurtStatusGetStarted()))
      || ((BurtStatusGetRunState() & STATEMASK) == COMPLETE))
  { 
    return(1);
  }

  return(0);
}


/*
 *----------------------------------------------------------------------
 *
 * LoadNextBackupTape --
 *
 *	This procedure prompts the user (via the logChannel) to mount a
 *      new tape,   Before doing so it rewinds and unmounts the old one.  
 *      This procedure blocks until it gets a new tape and is able to write 
 *      a tape label. 
 *
 * Results:
 *	
 *      TCL_OK if the tape was swapped and a new tape label was
 *      written, TCL_ERROR otherwise.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */

static int 
LoadNextBackupTape(Tcl_Channel mtChannel, Tcl_Channel logChannel)
{
  int fd;               /* Unix level file descriptor. */
  int foundEvent;
  char logBuffer[1024];
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch searchPtr;
  Burt_BackupControlBlock *backupCB;
  Burt_LoadNextTapeData promptData;
  Burt_TapeLabel tapeLabel;
  
  /* SunOS pukes at 2GB file size limit.
   * Reset seek ptr it will see the new tape. 
   */ 
  Tcl_GetChannelHandle(mtChannel, TCL_READABLE, (ClientData *)&fd);
  lseek(fd, 0, SEEK_SET);
  
  /* Print a pretty message about the EOT */
  sprintf(logBuffer, "BURT: End of tape %u in backup, "
	  "wrote %1.2f megabytes [%s]\n", 
	  BurtStatusGetTapes(), BurtStatusGetTotalXBytes(), BurtTime());
  Tcl_Write(logChannel, logBuffer, -1);
  Tcl_Flush(logChannel);
  
  /* Temporarily disable events on the backup channels */
  entryPtr = Tcl_FirstHashEntry(&backupCBTable, &searchPtr);
  while (entryPtr != NULL)
  {
    backupCB = (Burt_BackupControlBlock *)Tcl_GetHashValue(entryPtr);
    Tcl_CreateChannelHandler(backupCB->inputChannel, 0, 
			     ProcessBackupInputChannel, backupCB);
    Tcl_CreateChannelHandler(backupCB->stderrChannel, 0, 
			     ProcessBackupStderrChannel, backupCB); 
    Tcl_DeleteTimerHandler(backupCB->sentryTimer);
    
    entryPtr = Tcl_NextHashEntry(&searchPtr);
  }

  if (BurtMtSafeRewind(mtChannel) == MT_ERROR) 
    return(TCL_ERROR);
  
  /* This is the magic that makes partial readchecks happen.
   * First check the tape label on the tape in the drive.  This is a paranoia
   * check, it should be impossible to get here unless the labels match, but...
   */

  if (BurtMtReadLabel(mtChannel, &tapeLabel) != MT_OK)
    return(TCL_ERROR);

  sprintf(logBuffer, "%u", BurtStatusGetTapes());
  if ((strncmp(tapeLabel.label, BurtStatusGetTapeLabel(), 
	       BURT_MAX_LABELLEN) != 0) || 
      (strncmp(tapeLabel.tape, logBuffer, 3) != 0))
  {
    sprintf(logBuffer, "BURT: Wrong tape label in backup.  Expected tape #%d "
	    "from session %s, got tape #%s from session %s [%s]\n", 
	    BurtStatusGetTapes(), BurtStatusGetTapeLabel(), tapeLabel.tape, 
	    tapeLabel.label, BurtTime());
    Tcl_Write(logChannel, logBuffer, -1);
    Tcl_Flush(logChannel);
    return(TCL_ERROR);
  }

  /* Good, the labels match.  Now, save the backup status info */
  savedBackupStatus = BurtStatusSave();

  /* Then rewind and do readchecks */
  if (BurtMtSafeRewind(mtChannel) == MT_ERROR) 
    return(TCL_ERROR);
 
  if (BurtReadTape(backupMCB.interp, mtChannel, logChannel, 0) != TCL_OK)
  {
    /* BurtReadTape already filled in an error for us, we just write it out */
    Tcl_Write(logChannel, Tcl_GetStringResult(backupMCB.interp), -1);
    Tcl_Write(logChannel, "\n", -1);
    Tcl_Flush(logChannel);
    /* Make sure to restore the backup status stuff */
    BurtStatusRestore(savedBackupStatus);
    BurtStatusFree(savedBackupStatus);
    return(TCL_ERROR);
  }

  /* Restore the saved backup status */
  BurtStatusRestore(savedBackupStatus);
  BurtStatusFree(savedBackupStatus);

  /* Increment the tape count */
  BurtStatusSetTapes(BurtStatusGetTapes() + 1);
  
  /* SunOS dies at 2GB file size limit. Reset seek ptr to see new tape */ 
  Tcl_GetChannelHandle(mtChannel, TCL_READABLE, (ClientData *)&fd);
  lseek(fd, 0, SEEK_SET);
  
  /* Rewind and eject the tape */
  if (BurtMtSafeRewind(mtChannel) == MT_ERROR) 
    return(TCL_ERROR);

  if (BurtMtSafeOffline(mtChannel) == MT_ERROR) 
    return(TCL_ERROR);
  
  BurtStatusSetRunState(BACKUP | WAITING);

  /* set up promptData with the things the timer event handler will need */
  promptData.mtChannel = mtChannel;
  promptData.logChannel = logChannel;
  promptData.tapeReady = 0;

  /* Print initial prompt */
  sprintf(logBuffer, "BURT: Please mount tape %u (%d seconds) in backup... "
	  "[%s]\n", BurtStatusGetTapes(), TAPE_LOAD_DELAY, BurtTime());
  Tcl_Write(logChannel, logBuffer, -1);
  Tcl_Flush(logChannel);

  /* Create timer event which will check for the new tape */
  promptData.promptTimer = Tcl_CreateTimerHandler(TAPE_LOAD_DELAY * 1000,
						  ProcessLoadBackupTapeTimer,
						  (ClientData) &promptData);
  /* ProcessLoadBackupTapeTimer sets tapeReady to TRUE, so loop until that */
  foundEvent = 1;
  while (!promptData.tapeReady && foundEvent)
    foundEvent = Tcl_DoOneEvent(0);

  Tcl_DeleteTimerHandler(promptData.promptTimer);

  sprintf(logBuffer, "BURT: Mounted tape %u in backup.  "
	  "Have a nice day [%s]\n", BurtStatusGetTapes(), BurtTime());
  Tcl_Write(logChannel, logBuffer, -1);
  Tcl_Flush(logChannel);
  BurtStatusSetRunState(BACKUP | RUN);

  /* Re-enable the events on the backup channels */
  entryPtr = Tcl_FirstHashEntry(&backupCBTable, &searchPtr);
  while (entryPtr != NULL)
  {
    backupCB = (Burt_BackupControlBlock *)Tcl_GetHashValue(entryPtr);
    Tcl_CreateChannelHandler(backupCB->inputChannel, TCL_READABLE, 
			     ProcessBackupInputChannel, backupCB);
    Tcl_CreateChannelHandler(backupCB->stderrChannel, TCL_READABLE, 
			     ProcessBackupStderrChannel, backupCB); 
    backupCB->sentryTimer = Tcl_CreateTimerHandler(BURT_SENTRY_TIMEOUT, 
						   ProcessBackupSentry, 
						   (ClientData) backupCB);
    entryPtr = Tcl_NextHashEntry(&searchPtr);
  }
  
  return(TCL_OK);
}


/*
 *----------------------------------------------------------------------
 *
 * ProcessLoadBackupTapeTimer --
 *
 *	This function processes the timer events created to check for a
 *      new tape in the drive, and prompt the user for a tape if one is
 *      not ready.
 *
 * Results:
 *	
 * Side effects:
 *      
 *      A flag variable, tapeReady, may be set to 1 if a tape is ready.
 *      Also, a new timer event will be created if a tape is not ready.
 *      
 *----------------------------------------------------------------------
 */

static void
ProcessLoadBackupTapeTimer(ClientData clientData)
{
  Burt_LoadNextTapeData *data = (Burt_LoadNextTapeData *)clientData;
  Burt_TapeLabel tapeLabel;
  char logBuffer[1024];
  
  SetBurtTapeLabel(&tapeLabel);
  
  if ((BurtMtRewind(data->mtChannel) == MT_ERROR) ||
      (BurtMtWriteLabel(data->mtChannel, &tapeLabel) == MT_ERROR))
  {
    sprintf(logBuffer, "BURT: Please mount tape %u (%d seconds) in backup... "
	    "[%s]\n", BurtStatusGetTapes(), TAPE_LOAD_DELAY, BurtTime());
    Tcl_Write(data->logChannel, logBuffer, -1);
    Tcl_Flush(data->logChannel);
    data->promptTimer = Tcl_CreateTimerHandler(TAPE_LOAD_DELAY * 1000, 
					       ProcessLoadBackupTapeTimer,
					       (ClientData) data);
  }
  else
  {
    data->tapeReady = 1;
  }
  return;
}


/*
 *----------------------------------------------------------------------
 *
 * SetBurtTapeLabel --
 *
 *	Just like the name says, this function sets up a Burt_TapeLabel
 *      structure with the current information.
 *
 * Results:
 *	
 * Side effects:
 *      
 *      The Burt_TapeLabel struct pointed to by tapeLabel will get its 
 *      members overwritten with the current information.
 *----------------------------------------------------------------------
 */

static void
SetBurtTapeLabel(Burt_TapeLabel *tapeLabel)
{
  time_t now;
  char *dateStr;
  
  tapeLabel->magic = BURT_MAGIC_NUMBER;
  strncpy(tapeLabel->label, BurtStatusGetTapeLabel(), BURT_MAX_LABELLEN);
  now = time((time_t *)NULL);
  dateStr = ctime(&now);
  dateStr[strlen(dateStr) - 1] = '\0';
  strncpy(tapeLabel->date, dateStr, BURT_MAX_DATELEN);
  sprintf(tapeLabel->tape, "%u", BurtStatusGetTapes());
  return;
}


/*
 *----------------------------------------------------------------------
 *
 * BurtBackupRunningList --
 *
 *      Creates a listing of all currently running atoms (ie, those that
 *      are active and are being written to tape NOW)
 *
 * Results:
 *	Returns a standard Tcl list of all the elements in backupCBTable. 
 *      Null if the table is empty.  The caller is responsible for the 
 *      freeing of memory.  NOTE: It returns this list in a Tcl_DString *
 *      not just a char *!
 *      
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */

Tcl_DString *
BurtBackupRunningList(void)
{
  Tcl_DString *runningList;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;
  Burt_BackupControlBlock *backupCB;
    
  runningList = (Tcl_DString *)malloc(sizeof(Tcl_DString));
  if (runningList != NULL) 
  {
    Tcl_DStringInit(runningList);
    if ((BurtStatusGetRunState() & STATEMASK) != ABORT &&
	(BurtStatusGetRunState() & STATEMASK) != COMPLETE)
    {
      entryPtr = Tcl_FirstHashEntry(&backupCBTable, &search);
      while (entryPtr != NULL) 
      {
	backupCB = (Burt_BackupControlBlock *)Tcl_GetHashValue(entryPtr);
	Tcl_DStringStartSublist(runningList);
	Tcl_DStringAppendElement(runningList, backupCB->packet.host);
	Tcl_DStringAppendElement(runningList, backupCB->packet.atom);
	Tcl_DStringAppendElement(runningList, backupCB->packet.level);
	Tcl_DStringAppendElement(runningList, backupCB->packet.type);
	Tcl_DStringEndSublist(runningList);
	entryPtr = Tcl_NextHashEntry(&search);
      }
    }
  }

  return(runningList);
}
