/*
 * burtMtIO.c
 *
 *    This file implements the mag tape interface for burt.
 *
 * Copyright (C) 1995 - 1998 Eric Melski and Dean Jansa
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice and this permission notice appear in all copies,
 * and that all support documentation is included with each copy of the software.
 * A reasonable copying fee may be charged for any distribution of this 
 * software, but no fee may be charged for the software itself. You may 
 * distribute this software in aggregate with other programs as part of a 
 * larger software distribution provided that this software is not advertised 
 * as a product of your own.  Eric Melski and Dean Jansa make no representations
 * about the suitability of this software for any purpose.  It is provided 
 * "as is" without express or implied warranty.
 *
 * ERIC MELSKI AND DEAN JANSA DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL ERIC MELSKI OR DEAN JANSA BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 */


#ifndef lint
static char rcsid[] = "$Id: burtMtIO.c,v 1.17 1998/08/18 20:39:42 jmelski Exp $";
#endif

#ifdef _POSIX_SOURCE
    #undef _POSIX_SOURCE
#endif
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/mtio.h>  
#include <netinet/in.h>
#include <signal.h>
#define _POSIX_SOURCE
#include <unistd.h>
#include <math.h>
#include <burtInt.h> 


static void HeaderToNetworkOrder(Burt_Packet *packetPtr);
static void HeaderToHostOrder(Burt_Packet *packetPtr);

/*
 *----------------------------------------------------------------------
 *
 * Burt_MtCmd --
 *
 *	This procedure implements the Tcl command mt.
 *
 * Results:
 *      Returns a standard Tcl completion code, and leaves an error
 *	message in interp->result if an error occurs.
 *
 * Side effects:
 *	
 *
 *----------------------------------------------------------------------
 */

int 
Burt_MtCmd(ClientData cdata, Tcl_Interp *interp, int argc, char **argv)
{
  Tcl_Channel mtChannel;
  int length;
  int mode;

  if (argc != 3)
  {
    Tcl_SetResult(interp, "wrong # args: should be \"mt option channel\"", 
		  TCL_STATIC);
    return TCL_ERROR;
  }

  length = strlen(argv[1]);
  mtChannel = Tcl_GetChannel(interp, argv[2], &mode);

  if (mtChannel == NULL)
  {
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp, "Unable to find channel named \"", argv[2], "\"",
		     (char *) NULL);
    return TCL_ERROR;
  }

  if ((mode & (TCL_READABLE | TCL_WRITABLE)) == 0)
  {
    Tcl_SetResult(interp, "mt channel wasn't opened for reading or writing", 
		  TCL_STATIC);
    return TCL_ERROR;
  }

  if (strncmp(argv[1], "offline", length) == 0)
  {
    if (BurtMtOffline(mtChannel) == MT_ERROR)
    {
      Tcl_SetResult(interp, "Unable to offline channel", TCL_STATIC);
      return TCL_ERROR;
    }
  }
  else if (strncmp(argv[1], "rewind", length) == 0)
  {
    if (BurtMtRewind(mtChannel) == MT_ERROR)
    {
      Tcl_SetResult(interp, "Unable to rewind channel", TCL_STATIC);
      return TCL_ERROR;
    }
  }
  else
  {
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp, "Unknown option \"", argv[1], 
		     "\": should be offline or rewind", (char *) NULL);
    return TCL_ERROR;
  }

  return TCL_OK;
}


/*
 *----------------------------------------------------------------------
 *
 * BurtMtWrite --
 *
 *	This procedure writes a packet containing header information and
 *      the backup data to tape.
 *
 * Results:
 *	
 *	Returns the number of bytes written (the size of the packet) if 
 *      all went well, otherwise MT_EOT upon end of tape, or MT_ERROR 
 *      if there was a write error.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */

int
BurtMtWrite(Tcl_Channel mtChannel, Burt_Packet *packetPtr)
{
  int fd;                /* Unix level file descriptor. */
  int nbytes;
  int totalBytes;        /* Total number of bytes written. */
  
  Tcl_GetChannelHandle(mtChannel, TCL_WRITABLE, (ClientData *) &fd);
  HeaderToNetworkOrder(packetPtr);
  nbytes = write(fd, (void *) packetPtr, sizeof(Burt_Packet));
  HeaderToHostOrder(packetPtr);
  
  /* There are a couple of cases to watch for on writes:
   * 1.  A zero write.  This means EOT.
   * 2.  A negative write.  This means a write error.
   * 3.  A short write.  If the next write is zero, this is LEOT.
   *     If the next write is negative, this means a write error.
   */

  if (nbytes == 0) 
    return (MT_EOT);
  else if (nbytes < 0)
    return (MT_EOT);
  
  if (nbytes != sizeof(Burt_Packet)) 
  {
    HeaderToNetworkOrder(packetPtr);
    nbytes = write(fd, (void *) packetPtr, sizeof(Burt_Packet));
    HeaderToHostOrder(packetPtr);
    if (nbytes == 0)
      return(MT_EOT);
    else if (nbytes != 0 && nbytes != sizeof(Burt_Packet))
      return(MT_ERROR);
    /* If the second write was OK, great.  This will probably never happen */
  }

  totalBytes = nbytes;
  return(totalBytes);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtMtRead --
 *
 *	This procedure reads both a packet header and the buffer containing
 *      the backup data from the tape.  It deals with all the details of 
 *      reading blocks from tape (such as coverting to host byte order).
 *
 * Results:
 *	
 *	Returns the number of bytes read (the length of the buffer AND
 *      the sizeof the header.) if all went well, MT_ERROR upon a read 
 *      error and MT_EOT if it reached the EOT.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */

int 
BurtMtRead(Tcl_Channel mtChannel, Burt_Packet *packetPtr)
{
  int fd;             /* Unix level file descriptor. */
  int nbytes;
  int totalBytes;     /* Total bytes read, (header and buffer) */

  Tcl_GetChannelHandle(mtChannel, TCL_READABLE, (ClientData *) &fd);

  nbytes = read(fd, (void *) packetPtr, sizeof(Burt_Packet));
  if (nbytes == 0) {
    /* 
     * Read a file marker (or EOF)
     * Re-read to get packet header. 
     */
    nbytes = read(fd, (void *) packetPtr, sizeof(Burt_Packet));
      
    if (nbytes == 0) {
      return(MT_EOT);
    } else if (nbytes < 0) {
      sleep(TAPE_LOAD_DELAY);
      nbytes = read(fd, (void *) packetPtr, sizeof(Burt_Packet));
      if (nbytes == 0) {
	return(MT_EOT);
      } else if (nbytes < 0) {
	sleep(TAPE_LOAD_DELAY);
	nbytes = read(fd, (void *) packetPtr, sizeof(Burt_Packet));
	if (nbytes == 0) {
	  return(MT_EOT);
	} else if (nbytes < 0) {
	  sleep(TAPE_LOAD_DELAY);
	  nbytes = read(fd, (void *) packetPtr, sizeof(Burt_Packet));
	  if (nbytes == 0) {
	    return(MT_EOT);
	  } else if (nbytes < 0) {
	    return(MT_ERROR);
	  }
	}
      }
    }

  } 

  if (nbytes != sizeof(Burt_Packet)) {
    return(MT_ERROR);
  }

  totalBytes = nbytes;
  HeaderToHostOrder(packetPtr);

  if (packetPtr->blockSize == LOGICAL_EOT)
    return(MT_EOT);

  return(totalBytes);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtMtWriteLabel --
 *
 *	This procedure writes a tape label onto the tape.
 *      It should be used on each new tape in a backup. 
 *
 * Results:
 *	
 *      MT_OK if the label was written, MT_ERROR otherwise.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */
 
int
BurtMtWriteLabel(Tcl_Channel mtChannel, Burt_TapeLabel *label)
{
  int fd;              /* Unix level file descriptor. */
  int nbytes;
	   
  Tcl_GetChannelHandle(mtChannel, TCL_WRITABLE, (ClientData *) &fd);

  /* 
   * Ick!  The MtIO layer shouldn't be doing this, but the user layer
   * shouldn't know that it needs to be done!  Argh...
   */
  label->magic = (int)htonl(label->magic);

  nbytes = write(fd, (void *) label, sizeof(Burt_TapeLabel));
  
  if (nbytes != sizeof(Burt_TapeLabel))
    return(MT_ERROR);
  return(MT_OK);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtMtReadLabel --
 *
 *	This procedure reads a tape label from the tape.  It copies the
 *      label data to returnLabel, unless 
 *      returnLabel is NULL.
 *
 * Results:
 *	
 *      MT_OK if the label was read, MT_ERROR otherwise.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */

int
BurtMtReadLabel(Tcl_Channel mtChannel, Burt_TapeLabel *returnLabel)
{
  int fd;              /* Unix level file descriptor. */
  Burt_TapeLabel label;
  int nbytes;
  
  Tcl_GetChannelHandle(mtChannel, TCL_READABLE, (ClientData *) &fd);

  nbytes = read(fd, (void *) &label, sizeof(Burt_TapeLabel));
  
  if (nbytes != sizeof(Burt_TapeLabel)) 
  {
    /* See if the label is the size of the old tape labels */
    if (nbytes != 72)
    {
      returnLabel = NULL;
      return(MT_ERROR);
    }
  }

  if (returnLabel != NULL)
  {
    memcpy((void *)returnLabel, (void *)&label, sizeof(Burt_TapeLabel));
    /* 
     * Ick!  The MtIO layer shouldn't be doing this, but the user layer
     * shouldn't know that it needs to be done!  Argh...
     */
    returnLabel->magic = (int)ntohl(returnLabel->magic);
  }

  return(MT_OK);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtMtFastForward --
 *
 *	This procedure reads "count" filemarks on the tape to "fast forward"
 *      over data.
 *
 * Results:
 *	
 *      MT_OK if the filemarks were read, MT_ERROR otherwise.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */
 
int
BurtMtFastForward(Tcl_Channel mtChannel, int count)
{
    int fd;              /* Unix level file descriptor. */
    struct mtop mt_command;

    Tcl_GetChannelHandle(mtChannel, TCL_READABLE, (ClientData *) &fd);
    mt_command.mt_op = MTFSF;
    mt_command.mt_count = count;
    if (ioctl(fd, MTIOCTOP, &mt_command) < 0) {
	return(MT_ERROR);
    }

    return(MT_OK);
}



/*
 *----------------------------------------------------------------------
 *
 * BurtMtWriteFilemark --
 *
 *	This procedure writes a filemark onto the tape.
 *
 * Results:
 *	
 *      MT_OK if the filemark was written, MT_ERROR otherwise.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */
 
int
BurtMtWriteFilemark(Tcl_Channel mtChannel)
{
    int fd;              /* Unix level file descriptor. */
    struct mtop mt_command;

    Tcl_GetChannelHandle(mtChannel, TCL_WRITABLE, (ClientData *) &fd);
    mt_command.mt_op = MTWEOF;
    mt_command.mt_count = 1;
    if (ioctl(fd, MTIOCTOP, &mt_command) < 0) {
	return(MT_ERROR);
    }

    return(MT_OK);
}



/*
 *----------------------------------------------------------------------
 *
 * BurtMtRewind --
 *
 *	This procedure rewinds tape.
 *
 * Results:
 *	
 *      MT_OK if the tape was rewound, MT_ERROR otherwise.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */
 
int
BurtMtRewind(Tcl_Channel mtChannel)
{
  int fd;              /* Unix level file descriptor. */
  struct mtop mt_command;
  Tcl_GetChannelHandle(mtChannel, TCL_READABLE, (ClientData *) &fd);
  mt_command.mt_op = MTREW;
  mt_command.mt_count = 1;
  if (ioctl(fd, MTIOCTOP, &mt_command) < 0)
    return (MT_ERROR);

  return(MT_OK);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtMtSafeRewind --
 *
 *      Wrapper around BurtMtRewind that handles multiple retries
 *      
 * Results:
 *      
 *      MT_ERROR if all the retries fail; MT_OK otherwise.
 *	
 * Side effects:
 *      
 *      Rewinds the tape
 *      
 *----------------------------------------------------------------------
 */

int
BurtMtSafeRewind(Tcl_Channel mtChannel)
{
  int retries = 0;
  while (retries < 5 && BurtMtRewind(mtChannel) == MT_ERROR)
    retries++;

  if (retries == 5)
    return (MT_ERROR);

  return (MT_OK);
}


/*
 *----------------------------------------------------------------------
 *
 * BurtMtOffline --
 *
 *	This procedure takes a tape offline.
 *
 * Results:
 *	
 *      MT_OK if the tape was taken offline, MT_ERROR otherwise.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */
 
int
BurtMtOffline(Tcl_Channel mtChannel)
{
    int fd;              /* Unix level file descriptor. */
    struct mtop mt_command;

    Tcl_GetChannelHandle(mtChannel, TCL_READABLE, (ClientData *) &fd);
    mt_command.mt_op = MTOFFL;
    mt_command.mt_count = 1;
    if (ioctl(fd, MTIOCTOP, &mt_command) < 0)
      return(MT_ERROR);

    return(MT_OK);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtMtSafeOffline --
 *
 *      Wrapper around BurtMtOffline that handles multiple retries
 *      
 * Results:
 *      
 *      MT_ERROR if all the retries fail; MT_OK otherwise.
 *	
 * Side effects:
 *      
 *      Offlines the tape
 *      
 *----------------------------------------------------------------------
 */

int BurtMtSafeOffline(Tcl_Channel mtChannel)
{
  int retries = 0;
  while (retries < 5 && BurtMtOffline(mtChannel) == MT_ERROR)
    retries++;

  if (retries == 5)
    return (MT_ERROR);

  return (MT_OK);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtMtStatus --
 *
 *	This procedure sets the status of a tape drive.
 *
 * Results:
 *	
 *      MT_OK if the tape status was set, MT_ERROR otherwise.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */
 
int
BurtMtStatus(Tcl_Channel mtChannel)
{
    int fd;              /* Unix level file descriptor. */
    struct mtop mt_command;

    Tcl_GetChannelHandle(mtChannel, TCL_READABLE, (ClientData *) &fd);
    mt_command.mt_op = MTNOP;
    mt_command.mt_count = 1;
    if (ioctl(fd, MTIOCTOP, &mt_command) < 0) {
	return(MT_ERROR);
    }

    return(MT_OK);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtMtEOM --
 *
 *	This procedure advances the tape to the end of media marker
 *
 * Results:
 *	
 *      MT_OK if the tape status was set, MT_ERROR otherwise.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */
 
int
BurtMtEOM(Tcl_Channel mtChannel)
{
    int fd;              /* Unix level file descriptor. */
    struct mtop mt_command;

    Tcl_GetChannelHandle(mtChannel, TCL_READABLE, (ClientData *) &fd);
    mt_command.mt_op = MTEOM;
    mt_command.mt_count = 1;
    if (ioctl(fd, MTIOCTOP, &mt_command) < 0) {
	return(MT_ERROR);
    }

    return(MT_OK);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtMtBSF --
 *
 *	This procedure rewinds the tape count filemarks
 *
 * Results:
 *	
 *      MT_OK if the tape was rewound, MT_ERROR otherwise.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */
 
int
BurtMtBSF(Tcl_Channel mtChannel, int count)
{
    int fd;              /* Unix level file descriptor. */
    struct mtop mt_command;

    Tcl_GetChannelHandle(mtChannel, TCL_READABLE, (ClientData *) &fd);
    mt_command.mt_op = MTBSF;
    mt_command.mt_count = count;
    if (ioctl(fd, MTIOCTOP, &mt_command) < 0) {
	return(MT_ERROR);
    }

    return(MT_OK);
}


/*
 *----------------------------------------------------------------------
 *
 * HeaderToNetworkOrder --
 *
 *	This procedure converts a packet header from host byte order
 *      to network byte order.
 *
 * Results:
 *	
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */

static void 
HeaderToNetworkOrder(Burt_Packet *packetPtr)
{
    packetPtr->magic = (int)htonl(packetPtr->magic);
    packetPtr->blockNum = (int) htonl(packetPtr->blockNum);
    packetPtr->blockSize = (int) htonl(packetPtr->blockSize);
    packetPtr->totalSize = (unsigned int) htonl(packetPtr->totalSize);
    packetPtr->checksum = (unsigned int) htonl(packetPtr->checksum);

    return;
}



/*
 *----------------------------------------------------------------------
 *
 * HeaderToHostOrder --
 *
 *	This procedure converts a packet header from network byte order
 *      to the host byte order.
 *
 * Results:
 *	
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */

static void 
HeaderToHostOrder(Burt_Packet *packetPtr)
{
    packetPtr->magic = (int)ntohl(packetPtr->magic);
    packetPtr->blockNum = (int) ntohl(packetPtr->blockNum);
    packetPtr->blockSize = (int) ntohl(packetPtr->blockSize);
    packetPtr->totalSize = (unsigned int) ntohl(packetPtr->totalSize);
    packetPtr->checksum = (unsigned int) ntohl(packetPtr->checksum);

    return;
}







