/*
 * burtReadtape.c
 *
 *    This file implements the readtape command.  
 *
 * Copyright (C) 1995 - 1998 Eric Melski and Dean Jansa
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice and this permission notice appear in all copies,
 * and that all support documentation is included with each copy of the software.
 * A reasonable copying fee may be charged for any distribution of this 
 * software, but no fee may be charged for the software itself. You may 
 * distribute this software in aggregate with other programs as part of a 
 * larger software distribution provided that this software is not advertised 
 * as a product of your own.  Eric Melski and Dean Jansa make no representations
 * about the suitability of this software for any purpose.  It is provided 
 * "as is" without express or implied warranty.
 *
 * ERIC MELSKI AND DEAN JANSA DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL ERIC MELSKI OR DEAN JANSA BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 */

#ifndef lint
static char rcsid[] = "$Id: burtReadtape.c,v 1.20 1998/10/08 15:40:38 jmelski Exp $";
#endif

#include <burtInt.h>

typedef struct Burt_TOCItem {
  char key[BURT_MAX_HOSTLEN+
	  BURT_MAX_ATOMLEN+
	  BURT_MAX_LEVELLEN+
	  BURT_MAX_TYPELEN+1];
  unsigned int tape;
  unsigned int offset;
} Burt_TOCItem;

static Tcl_HashTable burtTOC;
static Tcl_HashTable burtInitTOC;
static int startedReadtape;
static char *savedReadtapeStatus;
static char *BurtGetTapeTOC(void);
static int LoadNextReadTape(Tcl_Channel mtChannel, Tcl_Channel logChannel);
static void ProcessLoadReadtapeTapeTimer(ClientData clientData);

/*
 *----------------------------------------------------------------------
 *
 * Burt_ReadtapeCmd --
 *
 *	This procedure implements the Tcl command readtape.
 *
 * Results:
 *      Returns a standard Tcl completion code, and leaves an error
 *	message in interp->result if an error occurs.
 *
 * Side effects:
 *	
 *
 *----------------------------------------------------------------------
 */

int 
Burt_ReadTapeCmd(ClientData cdata, Tcl_Interp *interp, int argc, char **argv)
{
  Tcl_Channel inputChannel;
  Tcl_Channel logChannel;
  char *TOC;
  Burt_TapeLabel label;
  int mode;
  char logBuffer[1024];

  if (argc != 4) 
  {
    Tcl_SetResult(interp, "wrong # args: should be \"readtape option "
		  "inputchannel logchannel\"", TCL_STATIC);
    return(TCL_ERROR);
  }

  if (!((strcmp(argv[1], "toc") == 0) || (strcmp(argv[1], "label") == 0))) 
  {
    Tcl_AppendResult(interp, "bad option \"", argv[1],
		     "\": should be toc or label", (char *) NULL);
    return(TCL_ERROR);
  }

  inputChannel = Tcl_GetChannel(interp, argv[2], &mode);
  if (inputChannel == NULL) 
    /* Tcl_GetChannel fills in the error message for us. */
    return(TCL_ERROR);

  if ((mode & TCL_READABLE) == 0) 
  {
    Tcl_SetResult(interp, "input channel wasn't opened for reading", TCL_STATIC);
    Tcl_UnregisterChannel(interp, inputChannel);
    return(TCL_ERROR);
  }     

  logChannel = Tcl_GetChannel(interp, argv[3], &mode);
  if (logChannel == NULL)
    /* Tcl_GetChannel fills in the error message for us. */
    return(TCL_ERROR);

  if ((mode & TCL_WRITABLE) == 0) 
  {
    Tcl_SetResult(interp, "log channel wasn't opened for writing", TCL_STATIC);
    Tcl_UnregisterChannel(interp, logChannel);
    return(TCL_ERROR);
  }     

  if (BurtMtRewind(inputChannel) == MT_ERROR) 
  {
    Tcl_SetResult(interp, "Could not rewind tape", TCL_STATIC);
    return(TCL_ERROR);
  }

    
  BurtStatusReset();

  if (strcmp(argv[1], "toc") == 0) 
  {
    if (BurtReadTape(interp, inputChannel, logChannel, 1) == TCL_OK) {
      TOC = BurtGetTapeTOC();
      if (TOC == NULL) 
      {
	Tcl_SetResult(interp, "BURT: Could not allocate memory for TOC!\n", 
		      TCL_STATIC);
	BurtStatusSetRunState(READTAPE | ABORT);
	BurtStatusSetCompleteTime(time((time_t *) NULL));
	BurtReadtapeInit();
	return(TCL_ERROR);
      } 
      else 
      {
	Tcl_SetResult(interp, TOC, TCL_DYNAMIC);
	BurtStatusSetRunState(READTAPE | COMPLETE);
	BurtStatusSetCompleteTime(time((time_t *) NULL));
	BurtReadtapeInit();
	return(TCL_OK);
      }
    } 
    else 
    {
      sprintf(logBuffer, "BURT: Outputting partial TOC, %1.2f bytes, "
	      "%d salvagable items\n", BurtStatusGetTotalXBytes(), 
	      BurtTableLength(&burtTOC));
      Tcl_AppendResult(interp, logBuffer, (char *)NULL);
      TOC = BurtGetTapeTOC();
      if (TOC == NULL)
      {
	Tcl_AppendResult(interp, "BURT: Could not allocate memory for TOC!\n",
			 (char *)NULL);
      }
      else
      {
	Tcl_AppendResult(interp, TOC, (char *)NULL);
	free((void *)TOC);
      }

      BurtStatusSetRunState(READTAPE | ABORT);
      BurtStatusSetCompleteTime(time((time_t *) NULL));
      return(TCL_ERROR);
    }
  } 
  else if (strcmp(argv[1], "label") == 0)
  {
    if (BurtMtReadLabel(inputChannel, &label) != MT_OK)
    {
      Tcl_SetResult(interp, "Could not read tape label", TCL_STATIC);
      return(TCL_ERROR);
    }

    sprintf(logBuffer, "%s %s %d {%s}", label.label, label.tape, label.magic, 
	    label.date);
	    
    Tcl_AppendResult(interp, logBuffer, (char *) NULL);
    BurtStatusSetRunState(READTAPE | COMPLETE);
    BurtStatusSetCompleteTime(time((time_t *) NULL));
    return(TCL_OK);
  }
  return(TCL_OK);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtReadtapeInit --
 *
 *	This procedure simply sets the startedReadtape flag to zero.  The
 *      flag is used by readtape to determine if partial readchecks have 
 *      already been done.
 *
 * Results:
 *
 * Side effects:
 *
 *      Sets startedReadtape to zero.
 *      
 *----------------------------------------------------------------------
 */

void
BurtReadtapeInit(void)
{
  startedReadtape = 0;
  return;
}

/*
 *----------------------------------------------------------------------
 *
 * BurtReadTape --
 *
 *	This procedure reads a burt tape and validates the data contained
 *      on it.  In the process it builds a tape table of contents, which
 *      may be retrieved by calling BurtGetTapeTOC().
 *
 * Results:
 *
 *     TCL_OK if tape is OK, TCL_ERROR otherwise, interp-> result is
 *     filled in with the error if TCL_ERROR
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */

int 
BurtReadTape(Tcl_Interp *interp, Tcl_Channel inputChannel, 
	     Tcl_Channel logChannel, int readMultipleTapes)
{
  Burt_Packet packet;
  Burt_TOCItem *TOCItemPtr, *tempTOCItemPtr;
  Tcl_HashEntry *entryPtr;
  int mtStatus;
  int new;
  int tapeOffset = 0;
  char logBuffer[1024];
  Burt_TapeLabel label;
  
  if (BurtMtReadLabel(inputChannel, &label) != MT_OK)
  {
    Tcl_SetResult(interp, "Could not read tape label", TCL_STATIC);
    return(TCL_ERROR);
  }

  /* If we've already started readchecks, reload the status info for them */
  if (startedReadtape == 1)
  {
    BurtStatusRestore(savedReadtapeStatus);
    BurtStatusFree(savedReadtapeStatus);
    
    /* Make sure the tape we're looking at belongs to the same set at the 
     * previously viewed tape */
    if (strcmp(BurtStatusGetTapeLabel(), label.label) != 0)
    {
      Tcl_SetResult(interp, "Current tape is not from the correct backup set.",
		    TCL_STATIC);
      return(TCL_ERROR);
    }
  }
  else  /* startedReadtape == 0 */
  {
    Tcl_InitHashTable(&burtInitTOC, TCL_STRING_KEYS);
    Tcl_InitHashTable(&burtTOC, TCL_STRING_KEYS);
    
    BurtStatusReset();
    BurtStatusSetStartTime(time((time_t *) NULL));
  }

  BurtStatusSetRunState(READTAPE | RUN);
  BurtStatusSetTapeLabel(label.label);
  BurtStatusSetChannels(1);
  BurtStatusSetTapes((atoi(label.tape)));
  startedReadtape = 1;
  
  TOCItemPtr = (Burt_TOCItem *) malloc(sizeof(Burt_TOCItem));

  do 
  {
    while (Tcl_DoOneEvent(TCL_ALL_EVENTS|TCL_DONT_WAIT) != 0);

    mtStatus = BurtMtRead(inputChannel, &packet);

    if (mtStatus == MT_EOT) 
    {
      /* If we're only reading one tape right now, return */
      if (readMultipleTapes == 0) 
      {
	savedReadtapeStatus = BurtStatusSave();
	free((void *)TOCItemPtr);
	return (TCL_OK);
      }

      /* Get the next tape and get on with it already. */
      if (LoadNextReadTape(inputChannel, logChannel) == TCL_ERROR) 
      {
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp, 
			 "BURT: --------------------------------------\n",
			 "BURT: Readtape I/O error, couldn't load tape\n",
			 "BURT: --------------------------------------\n",
			 (char *) NULL);
	free ((void *)TOCItemPtr);
	return (TCL_ERROR);
      }
      continue;
    }
    else if (mtStatus == MT_ERROR) 
    {
      /* MT_ERROR *probably* means a short read of the buffer.
       * Try to read again; we may be at EOM, in which case BurtMtRead will
       * return EOT.  If it doesn't, then we are well and truly screwed.
       */
      mtStatus = BurtMtRead(inputChannel, &packet);
      if (mtStatus == MT_EOT) 
      {
	/* If we're only reading one tape right now, return */
	if (readMultipleTapes == 0) 
	{
	  savedReadtapeStatus = BurtStatusSave();
	  free((void *)TOCItemPtr);
	  return (TCL_OK);
	}

	if (LoadNextReadTape(inputChannel, logChannel) == TCL_ERROR) 
	{
	  Tcl_ResetResult(interp);
	  Tcl_AppendResult(interp, 
			   "BURT: --------------------------------------\n",
			   "BURT: Readtape I/O error, couldn't load tape\n",
			   "BURT: --------------------------------------\n",
			   (char *) NULL);
	  free ((void *)TOCItemPtr);
	  return (TCL_ERROR);
	}
	continue;
      } 
      else if (mtStatus == MT_ERROR) 
      {
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp, 
			 "BURT: ------------------------------------\n",
			 "BURT: Readtape read error, readchecks fail\n",
			 "BURT: ------------------------------------\n",
			 (char *) NULL);
	free ((void *)TOCItemPtr);
	return(TCL_ERROR);
      }
    }

    if (packet.blockSize != END_OF_BACKUPSET)
    {
      if ((BurtChecksum((void *) packet.buffer, 
			packet.blockSize) != packet.checksum)) 
      {
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp, 
			 "BURT: ------------------------\n",
			 "BURT: Readtape Checksum Error: ", packet.host, 
			 " ", packet.atom, " ", packet.level,
			 " ", packet.type, "\n"
			 "BURT: ------------------------\n",
			 (char *) NULL);
	free ((void *)TOCItemPtr);
	return(TCL_ERROR);
      }
      else /* checksums match */
      {
	/* add this atom to the TOC if this is a first block */
	if (packet.blockNum == 1) 
	{
	  tapeOffset++;
	  strcpy(TOCItemPtr->key, packet.host);
	  strcat(TOCItemPtr->key, " ");
	  strcat(TOCItemPtr->key, packet.atom);
	  strcat(TOCItemPtr->key, " ");
	  strcat(TOCItemPtr->key, packet.level);
	  strcat(TOCItemPtr->key, " ");
	  strcat(TOCItemPtr->key, packet.type);
	  TOCItemPtr->tape = BurtStatusGetTapes();
	  TOCItemPtr->offset = tapeOffset;
	  
	  entryPtr = Tcl_CreateHashEntry(&burtInitTOC, TOCItemPtr->key, &new);
	  
	  if (new) 
	  {
	    Tcl_SetHashValue(entryPtr, TOCItemPtr);
	    TOCItemPtr = (Burt_TOCItem *) malloc(sizeof(Burt_TOCItem));
	    BurtStatusSetStarted(BurtStatusGetStarted() + 1);
	    sprintf(logBuffer, "BURT: Created entry {{%s %s %s %s} %d %d} "
		    "[%s]\n",
		    packet.host, packet.atom, packet.level, packet.type,
		    BurtStatusGetTapes(), tapeOffset, BurtTime());
	    Tcl_Write(logChannel, logBuffer, strlen(logBuffer));
	    Tcl_Flush(logChannel);
	  } 
	}
	
	/* If this is the end of a backup, put out a message */
	if (packet.blockSize == END_OF_BACKUP)
	{
	  strcpy(TOCItemPtr->key, packet.host);
	  strcat(TOCItemPtr->key, " ");
	  strcat(TOCItemPtr->key, packet.atom);
	  strcat(TOCItemPtr->key, " ");
	  strcat(TOCItemPtr->key, packet.level);
	  strcat(TOCItemPtr->key, " ");
	  strcat(TOCItemPtr->key, packet.type);
	  
	  BurtStatusSetFinished(BurtStatusGetFinished() + 1);
	  
	  entryPtr = Tcl_FindHashEntry(&burtInitTOC, TOCItemPtr->key);
	  if (entryPtr != NULL)
	  {
	    tempTOCItemPtr = (Burt_TOCItem *)Tcl_GetHashValue(entryPtr);
	    Tcl_DeleteHashEntry(entryPtr);
	    
	    entryPtr = Tcl_CreateHashEntry(&burtTOC, tempTOCItemPtr->key, &new);
	    if (new)
	      Tcl_SetHashValue(entryPtr, tempTOCItemPtr);

	    sprintf(logBuffer, "BURT: Completed entry {{%s %s %s %s} %d %d} "
		    "[%s]\n", packet.host, packet.atom, packet.level, 
		    packet.type, tempTOCItemPtr->tape, tempTOCItemPtr->offset,
		    BurtTime());
	  }
	  else
	  {
	    sprintf(logBuffer, "BURT: Completed entry {{%s %s %s %s} ? ?} "
		    "[%s]\n", packet.host, packet.atom, packet.level, 
		    packet.type, BurtTime());
	  }
	  Tcl_Write(logChannel, logBuffer, strlen(logBuffer));
	  Tcl_Flush(logChannel);
	}

	/* if this is an aborted or killed dump, remove it from the TOC */
	if (packet.blockSize == ABORTED_BACKUP || 
	    packet.blockSize == KILLED_BACKUP)
	{
	  strcpy(TOCItemPtr->key, packet.host);
	  strcat(TOCItemPtr->key, " ");
	  strcat(TOCItemPtr->key, packet.atom);
	  strcat(TOCItemPtr->key, " ");
	  strcat(TOCItemPtr->key, packet.level);
	  strcat(TOCItemPtr->key, " ");
	  strcat(TOCItemPtr->key, packet.type);
	  
	  entryPtr = Tcl_FindHashEntry(&burtInitTOC, TOCItemPtr->key);
	  if (entryPtr != NULL)
	  {
	    BurtStatusSetFinished(BurtStatusGetFinished() + 1);
	    tempTOCItemPtr = (Burt_TOCItem *)Tcl_GetHashValue(entryPtr);
	    sprintf(logBuffer, "BURT: Removed entry {{%s %s %s %s} %d %d} "
		    "[%s]\n",
		    packet.host, packet.atom, packet.level, packet.type,
		    tempTOCItemPtr->tape, tempTOCItemPtr->offset, BurtTime());
	    Tcl_Write(logChannel, logBuffer, strlen(logBuffer));
	    Tcl_Flush(logChannel);
	    free((void *)tempTOCItemPtr);
	    Tcl_DeleteHashEntry(entryPtr);
	  }
	}
      }
    }
    BurtStatusSetBytes(BurtStatusGetBytes() + mtStatus);
  }
  while (packet.blockSize != END_OF_BACKUPSET);
    
  free((void *)TOCItemPtr);
  BurtStatusSetRunState(READTAPE | COMPLETE);
  BurtStatusSetCompleteTime(time((time_t *) NULL));
  return(TCL_OK);
}

/*
 *----------------------------------------------------------------------
 *
 * BurtGetTapeTOC --
 *
 *     This procedure returns the tape's TOC (as built by calling 
 *     BurtReadTape).  It is the callers responsibility to free the TOC 
 *     list it returns.
 *
 * Results:
 *      The tape's Table of Contents (if there was one created) or NULL.    
 *     
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */

static char *
BurtGetTapeTOC(void)
{
  Tcl_DString TOCList;
  Burt_TOCItem *TOCItemPtr;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;
  char buffer[64];
  char *TOCStr;

  Tcl_DStringInit(&TOCList);
  entryPtr = Tcl_FirstHashEntry(&burtTOC, &search);
  while (entryPtr != NULL) 
  {
    TOCItemPtr = (Burt_TOCItem *)Tcl_GetHashValue(entryPtr);
    Tcl_DStringStartSublist(&TOCList);
    Tcl_DStringAppendElement(&TOCList, TOCItemPtr->key);
    sprintf(buffer, "%d", TOCItemPtr->tape);
    Tcl_DStringAppendElement(&TOCList, buffer); 
    sprintf(buffer, "%d", TOCItemPtr->offset);
    Tcl_DStringAppendElement(&TOCList, buffer); 
    Tcl_DStringEndSublist(&TOCList);
    entryPtr = Tcl_NextHashEntry(&search);
  }

  TOCStr = (char *)malloc(Tcl_DStringLength(&TOCList) + 11);
  sprintf(TOCStr, "BURT: TOC %s", Tcl_DStringValue(&TOCList));
  
  Tcl_DStringFree(&TOCList);
  return(TOCStr);
}

/*
 *----------------------------------------------------------------------
 *
 * LoadNextReadTape --
 *
 *	This procedure prompts the user (via the logChannel) to mount the
 *      next tape, Before doing so it rewinds and unmounts the old one.  
 *      This procedure blocks until it gets a new tape and is able to read 
 *      a tape label. 
 *
 * Results:
 *	
 *      TCL_OK if the tape was swapped and the next tape label was
 *      read.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */

static int 
LoadNextReadTape(Tcl_Channel mtChannel, Tcl_Channel logChannel)
{
  int fd;               /* Unix level file descriptor. */
  char logBuffer[1024]; /* Buffer for pretty log output */
  int foundEvent;
  Burt_LoadNextTapeData promptData;

  if (BurtMtSafeRewind(mtChannel) == MT_ERROR) 
    return (TCL_ERROR);
  
  if (BurtMtSafeOffline(mtChannel) == MT_ERROR) 
    return (TCL_ERROR);

  /*
   * SunOS pukes at 2GB.  Reset the file seek pointer
   */
  Tcl_GetChannelHandle(mtChannel, TCL_READABLE, (ClientData *) &fd);
  lseek(fd, 0, SEEK_SET);
  
  sprintf(logBuffer, "BURT: End of tape %u in readtape, read %1.2f megabytes "
	  "[%s]\n", BurtStatusGetTapes(), BurtStatusGetTotalXBytes(), 
	  BurtTime());
  Tcl_Write(logChannel, logBuffer, strlen(logBuffer));
  Tcl_Flush(logChannel);
  
  /* Increment the tape count */
  BurtStatusSetTapes(BurtStatusGetTapes() + 1);
  
  /* set up the Burt_LoadNextTapeData structure */
  promptData.mtChannel = mtChannel;
  promptData.logChannel = logChannel;
  promptData.tapeReady = 0;
  
  sprintf(logBuffer, "BURT: Please mount tape %u (%d seconds) in readtape... "
	  "[%s]\n", BurtStatusGetTapes(), TAPE_LOAD_DELAY, BurtTime());
  Tcl_Write(logChannel, logBuffer, -1);
  Tcl_Flush(logChannel);
  
  BurtStatusSetRunState(READTAPE | WAITING);
  promptData.promptTimer = Tcl_CreateTimerHandler(TAPE_LOAD_DELAY * 1000,
						  ProcessLoadReadtapeTapeTimer,
						  (ClientData) &promptData);
  foundEvent = 1;
  while (!promptData.tapeReady && foundEvent)
    foundEvent = Tcl_DoOneEvent(0);

  Tcl_DeleteTimerHandler(promptData.promptTimer);

  BurtStatusSetRunState(READTAPE | RUN);
  sprintf(logBuffer, "BURT: Mounted tape %u in readtape.  "
	  "Thank you, have a nice day [%s]\n", BurtStatusGetTapes(), BurtTime());
  Tcl_Write(logChannel, logBuffer, -1);
  Tcl_Flush(logChannel);

  return(TCL_OK);
}

static void
ProcessLoadReadtapeTapeTimer(ClientData clientData)
{
  Burt_LoadNextTapeData *data = (Burt_LoadNextTapeData *)clientData;
  Burt_TapeLabel tapeLabel;
  char logBuffer[1024];
  char tapeNo[3];
  char *label;

  if ((BurtMtRewind(data->mtChannel) == MT_ERROR) ||
      (BurtMtReadLabel(data->mtChannel, &tapeLabel) == MT_ERROR))
  {
    sprintf(logBuffer, "BURT: Please mount tape %u (%d seconds) in readtape... "
	    "[%s]\n", BurtStatusGetTapes(), TAPE_LOAD_DELAY, BurtTime());
    Tcl_Write(data->logChannel, logBuffer, -1);
    Tcl_Flush(data->logChannel);
    data->promptTimer = Tcl_CreateTimerHandler(TAPE_LOAD_DELAY * 1000, 
					       ProcessLoadReadtapeTapeTimer,
					       clientData);
  }
  else
  {
    /* Make sure they put the right tape in the drive! */
    sprintf(tapeNo, "%u", BurtStatusGetTapes());
    label = BurtStatusGetTapeLabel();
    if ((strncmp(tapeLabel.label, label, BURT_MAX_LABELLEN) != 0) || 
	(strncmp(tapeLabel.tape, tapeNo, 3) != 0))
    {
      sprintf(logBuffer, "BURT: Wrong tape in readtape.  Expected tape #%s "
	      "from session %s, got tape #%s from session %s [%s]\n", 
	      tapeNo, label, tapeLabel.tape, tapeLabel.label, BurtTime());
      Tcl_Write(data->logChannel, logBuffer, -1);
      Tcl_Flush(data->logChannel);
      data->promptTimer = Tcl_CreateTimerHandler(TAPE_LOAD_DELAY * 1000, 
						 ProcessLoadReadtapeTapeTimer,
						 clientData);
      BurtMtOffline(data->mtChannel);
    }
    else
    {
      data->tapeReady = 1;
    }
  }
  return;
}

/*
 *----------------------------------------------------------------------
 *
 * BurtReadtapeRunningList --
 *
 *      Creates a listing of all currently running atoms (ie, those that
 *      are active and are being written to tape NOW)
 *
 * Results:
 *	Returns a standard Tcl list of all the elements in backupCBTable. 
 *      Null if the table is empty.  The caller is responsible for the 
 *      freeing of memory.  NOTE: It returns this list in a Tcl_DString *
 *      not just a char *!
 *      
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */

Tcl_DString *
BurtReadtapeRunningList(void)
{
  Tcl_DString *runningList;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;
  Burt_TOCItem *TOCItemPtr;
  
  runningList = (Tcl_DString *)malloc(sizeof(Tcl_DString));
  if (runningList != NULL) 
  {
    Tcl_DStringInit(runningList);
    entryPtr = Tcl_FirstHashEntry(&burtInitTOC, &search);
    while (entryPtr != NULL) 
    {
      TOCItemPtr = (Burt_TOCItem *)Tcl_GetHashValue(entryPtr);
      Tcl_DStringAppendElement(runningList, TOCItemPtr->key);
      entryPtr = Tcl_NextHashEntry(&search);
    }
  }

  return(runningList);
}
