/*
 * burtRecover.c
 *
 *    This file implements the recover command.  This is the nuts and bolts
 *    of Burt's recover services.
 *
 * Copyright (C) 1995 - 1998 Eric Melski and Dean Jansa
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice and this permission notice appear in all copies,
 * and that all support documentation is included with each copy of the software.
 * A reasonable copying fee may be charged for any distribution of this 
 * software, but no fee may be charged for the software itself. You may 
 * distribute this software in aggregate with other programs as part of a 
 * larger software distribution provided that this software is not advertised 
 * as a product of your own.  Eric Melski and Dean Jansa make no representations
 * about the suitability of this software for any purpose.  It is provided 
 * "as is" without express or implied warranty.
 *
 * ERIC MELSKI AND DEAN JANSA DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL ERIC MELSKI OR DEAN JANSA BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 */

#ifndef lint
static char rcsid[] = "$Id: burtRecover.c,v 1.17 1998/08/18 20:39:43 jmelski Exp $";
#endif

#include <burtInt.h>
#include <errno.h>

/* the Burt_RecoverControlBlock holds information specific to each atom */
typedef struct Burt_RecoverControlBlock {
  Tcl_Channel outputChannel;
  char host[BURT_MAX_HOSTLEN];
  char atom[BURT_MAX_ATOMLEN];
  char type[BURT_MAX_TYPELEN];
  char level[BURT_MAX_LEVELLEN];
  unsigned int lastBlock;
  unsigned int totalSize;
} Burt_RecoverControlBlock;

/* the Burt_RecoverMetaControlBlock holds information common to all atoms */
typedef struct Burt_RecoverMetaControlBlock {
  Tcl_Interp *interp;
  Tcl_Channel inputChannel;
  Tcl_Channel logChannel;
} Burt_RecoverMetaControlBlock;

static int StartRecoverChannel(Tcl_Interp *interp, Tcl_Channel inputChannel, 
			       Tcl_Channel logChannel);
static void ProcessRecoverInputChannel(Burt_RecoverMetaControlBlock *recoverMCB);
static int GetUserRecoverChannel(Tcl_Interp *interp, 
				 Burt_RecoverControlBlock *recoverCB);
static int BurtRecoverIsComplete(void);
static int LoadNextRecoverTape(Tcl_Channel mtChannel, Tcl_Channel logChannel);
static void ProcessLoadRecoverTapeTimer(ClientData clientData);

static Tcl_HashTable recoverCBTable;

/*
 *----------------------------------------------------------------------
 *
 * Burt_RecoverCmd --
 *
 *	This procedure implements the Tcl command recover.
 *
 * Results:
 *      Returns a standard Tcl completion code, and leaves an error
 *	message in interp->result if an error occurs.
 *
 * Side effects:
 *	
 *
 *----------------------------------------------------------------------
 */

int 
Burt_RecoverCmd(ClientData cdata, Tcl_Interp *interp, int argc, char **argv)
{
  Tcl_Channel inputChannel;
  Tcl_Channel logChannel;
  Burt_TapeLabel label;
  int filemarkCount;
  int mode;
  int length;
  char logBuffer[1024];

  if ((argc < 2) || (argc > 5))
  {
    Tcl_SetResult(interp, "wrong # args: should be \"recover option ?arg?\"",
		  TCL_STATIC);
    return TCL_ERROR;
  }
      
  length = strlen(argv[1]);
  if (strncmp(argv[1], "start", length) == 0)
  {
    if ((argc < 4) || (argc > 5)) 
    {
      Tcl_AppendResult(interp, "wrong # args: should be ", 
		       "\"recover start inputchannel logchannel ",
		       "?skipfilemarks?\"", (char *) NULL);
      return TCL_ERROR;
    }
    
    if (BurtTableLength(&scheduleTable) == 0)
    {
      Tcl_SetResult(interp, "Nothing scheduled for recover.  Aborting",
		    TCL_STATIC);
      return(TCL_ERROR);
    }

    inputChannel = Tcl_GetChannel(interp, argv[2], &mode);
    if( inputChannel == NULL ) 
      return(TCL_ERROR);

    if ((mode & TCL_READABLE) == 0) 
    {
      Tcl_UnregisterChannel(interp, inputChannel);
      Tcl_SetResult(interp, "input channel wasn't opened for reading",
		    TCL_STATIC);
      return(TCL_ERROR);
    }     

    logChannel = Tcl_GetChannel(interp, argv[3], &mode);
    if( logChannel == NULL ) 
      return(TCL_ERROR);
    
    if ((mode & TCL_WRITABLE) == 0) 
    {
      Tcl_UnregisterChannel(interp, logChannel);
      Tcl_SetResult(interp, "log channel wasn't opened for writing", TCL_STATIC);
      return(TCL_ERROR);
    }     

    if (BurtMtRewind(inputChannel) == MT_ERROR) {
      Tcl_SetResult(interp, "Could not rewind tape", TCL_STATIC);
      return(TCL_ERROR);
    }

    if (BurtMtReadLabel(inputChannel, &label) != MT_OK)
    {
      Tcl_SetResult(interp, "Could not read tape label", TCL_STATIC);
      return(TCL_ERROR);
    }
    
    BurtStatusReset();
    BurtStatusSetRunState(RECOVER | RUN);
    BurtStatusSetScheduled(BurtTableLength(&scheduleTable));
    BurtStatusSetTapeLabel(label.label);
    BurtStatusSetChannels(1);
    BurtStatusSetFinished(0);
    BurtStatusSetStarted(0);
    BurtStatusSetTapes((atoi(label.tape)));
    BurtStatusSetBytes(0);
    BurtStatusSetStartTime(time((time_t *) NULL));

    Tcl_InitHashTable(&recoverCBTable, TCL_STRING_KEYS);

    if (argc > 4)
    {
      if (Tcl_GetInt(interp, argv[4], &filemarkCount) != TCL_OK)
	return TCL_ERROR;

      sprintf(logBuffer, "BURT: Attempting to skip %d filemarks... [%s]\n", 
	      filemarkCount, BurtTime());
      Tcl_Write(logChannel, logBuffer, strlen(logBuffer));
      Tcl_Flush(logChannel);
      if (BurtMtFastForward(inputChannel, filemarkCount) == MT_ERROR) 
      {
	Tcl_UnregisterChannel(interp, inputChannel);
	sprintf(logBuffer, "BURT: Fast Forward failed, try again without "
		"\"skipfilemarks\" [%s]", BurtTime());
	Tcl_AppendResult(interp, logBuffer, (char *) NULL);
	Tcl_Write(logChannel, logBuffer, strlen(logBuffer));
	Tcl_Flush(logChannel);
	return(TCL_ERROR);
      }
      sprintf(logBuffer, "BURT: Successfully skipped %d filemarks [%s]\n", 
	      filemarkCount, BurtTime());
      Tcl_Write(logChannel, logBuffer, strlen(logBuffer));
      Tcl_Flush(logChannel);
    }

    return (StartRecoverChannel(interp, inputChannel, logChannel));
  }
  else
  {
    Tcl_AppendResult(interp, "bad option \"", argv[1], "\": should be start", 
		     (char *) NULL);
    return TCL_ERROR;
  }
}

/*
 *----------------------------------------------------------------------
 *
 * StartRecoverChannel --
 *
 *	This procedure sets up the Burt_RecoverControlBlock hash table with 
 *      the atoms that were scheduled for recovery from the tape.
 *
 * Results:
 *      Returns TCL_ERROR on error, TCL_OK otherwise
 *
 * Side effects:
 *	Creates an array of Burt_RecoverControlBlocks
 *
 *----------------------------------------------------------------------
 */

static int
StartRecoverChannel(Tcl_Interp *interp, Tcl_Channel inputChannel, 
		    Tcl_Channel logChannel)
{
  Burt_ScheduleItem *scheduleItemPtr;
  Burt_RecoverControlBlock *recoverCB;
  Burt_RecoverMetaControlBlock recoverMCB;
  Tcl_HashEntry *entryPtr;
  char key[BURT_MAX_HOSTLEN + BURT_MAX_ATOMLEN + 
	  BURT_MAX_LEVELLEN + BURT_MAX_TYPELEN + 1];
  int new;

  /* Set up the Burt_RecoverMetaControlBlock with common information */
  recoverMCB.interp = interp;
  recoverMCB.inputChannel = inputChannel;
  recoverMCB.logChannel = logChannel;

  /* allocate and initialize each recoverCB; put it in the hash table */
  scheduleItemPtr = BurtScheduleGetNext(recoverMCB.interp, recoverMCB.logChannel,
					&scheduleTable);
  while (scheduleItemPtr != NULL)
  {
    recoverCB = (Burt_RecoverControlBlock *)
      malloc(sizeof(Burt_RecoverControlBlock));
    if (recoverCB == NULL) 
    {
      Tcl_SetResult(interp, "Out of Memory in StartRecoverChannel", TCL_STATIC);
      return TCL_ERROR;
    }
  
    recoverCB->outputChannel = (Tcl_Channel) NULL;
    recoverCB->lastBlock = 0;
    recoverCB->totalSize = 0;
    strncpy(recoverCB->host, scheduleItemPtr->host, BURT_MAX_HOSTLEN);
    strncpy(recoverCB->atom, scheduleItemPtr->atom, BURT_MAX_ATOMLEN);
    strncpy(recoverCB->level, scheduleItemPtr->level, BURT_MAX_LEVELLEN);
    strncpy(recoverCB->type, scheduleItemPtr->type, BURT_MAX_TYPELEN);
    strncpy(key, scheduleItemPtr->host, BURT_MAX_HOSTLEN);
    strcat(key, scheduleItemPtr->atom);
    strcat(key, scheduleItemPtr->level);
    strcat(key, scheduleItemPtr->type);

    entryPtr = Tcl_CreateHashEntry(&recoverCBTable, key, &new);
    if (new)
      Tcl_SetHashValue(entryPtr, recoverCB);
    free((void *) scheduleItemPtr);
    scheduleItemPtr = BurtScheduleGetNext(recoverMCB.interp, 
					  recoverMCB.logChannel, &scheduleTable);
  }
  
  ProcessRecoverInputChannel(&recoverMCB);
  return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * GetUserRecoverChannel --
 *
 *	This procedure fills in the outputChannel field of the recover 
 *      control block.  It does this by calling a user defined TCL procedure 
 *      (looked up in a Tcl array based on the "type" field of the recover
 *      control block.) which must return one writeable channel.  (This is 
 *      where the data which ends up going to the user defined proc comes 
 *      from).
 *
 * Results:
 *	Returns TCL_OK if the channels was gathered, TCL_ERROR otherwise.
 *      If TCL_ERROR is returned, the interp->result field is filled
 *      in with an error message.
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

int
GetUserRecoverChannel(Tcl_Interp *interp, Burt_RecoverControlBlock *recoverCB)
{
  char *userProc;
  int mode;
  char *outputChannelStr;
  
  userProc = Tcl_GetVar2(interp, "burt_recover_proc", recoverCB->type, 
			 TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
  if (userProc == NULL) 
  {
    /*
     * Tcl_GetVar2 already filled in the error, couldn't find
     * the user defined procedure. 
     */
    return(TCL_ERROR);
  }

  if (Tcl_VarEval(interp, userProc, " ", recoverCB->host, " ",
		  recoverCB->atom, " ", recoverCB->level, NULL) == TCL_ERROR) 
  {
    /* 
     * Tcl_VarEval found an error in user proc and filled in the error.
     */
    return(TCL_ERROR);
  }

  outputChannelStr = (char *) malloc(strlen(interp->result) + 1);
  if (outputChannelStr == NULL) 
  {
    Tcl_SetResult(interp, "Out of memory in GetUserRecoverChannel", TCL_STATIC);
    return(TCL_ERROR);
  } 
  else
  {
    strcpy(outputChannelStr, interp->result);
  }

  recoverCB->outputChannel = Tcl_GetChannel(interp, outputChannelStr, &mode);
  if (recoverCB->outputChannel == NULL) 
  {
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp, "can not find channel named \"", outputChannelStr,
		     "\" in user defined procedure", userProc, (char *) NULL);
    free((void *) outputChannelStr);
    return(TCL_ERROR);
  }

  if ((mode & TCL_WRITABLE) == 0) 
  {
    Tcl_ResetResult(interp);
    Tcl_AppendResult(interp, "channel ", outputChannelStr,
		     " wasn't opened for writing in user defined procedure ", 
		     userProc, (char *) NULL);
    Tcl_UnregisterChannel(interp, recoverCB->outputChannel);
    free((void *) outputChannelStr);
    return(TCL_ERROR);
  }     

  free((void *) outputChannelStr);
  return(TCL_OK);
}

/*
 *----------------------------------------------------------------------
 *
 * ProcessRecoverInputChannel --
 *
 *	This procedure processes input from (usually a tape drive)
 *      the inputChannel.  It compares the packet header read from the
 *      channel to the data in the RecoverControlBlock, if it is a match
 *      it sends the buffer to the outputChannel (which the user supplied
 *      via a Tcl procedure.)
 *
 * Results:
 *
 *
 * Side effects:
 *	
 *
 *----------------------------------------------------------------------
 */

static void ProcessRecoverInputChannel(Burt_RecoverMetaControlBlock *recoverMCB)
{
  Burt_Packet packet;
  Tcl_HashEntry *entryPtr;
  Burt_RecoverControlBlock *recoverCB;
  char logBuffer[1024];
  char key[BURT_MAX_HOSTLEN + BURT_MAX_ATOMLEN + 
	  BURT_MAX_LEVELLEN + BURT_MAX_TYPELEN + 1];
  int nbytes;
  
  /* 
   * Read headers until we find a block of a dump we are looking for
   */
  do 
  {
    /* Let some other things happen */
    while (Tcl_DoOneEvent(TCL_ALL_EVENTS | TCL_DONT_WAIT) != 0);
    nbytes = BurtMtRead(recoverMCB->inputChannel, &packet);

    if (nbytes == MT_EOT) 
    {
      /* Get the next tape and get on with it already. */
      if (LoadNextRecoverTape(recoverMCB->inputChannel, 
			      recoverMCB->logChannel) == TCL_ERROR) 
      {
	sprintf(logBuffer, "BURT: I/O error during recover, couldn't load "
		"next tape [%s]\n", BurtTime());
	Tcl_Write(recoverMCB->logChannel, logBuffer, strlen(logBuffer));
	Tcl_Flush(recoverMCB->logChannel);
	return;
      }
      continue;
    } 
    else if (nbytes == MT_ERROR) 
    {
      /* MT_ERROR probably means a short read of the buffer */
/*      if (packet.blockSize == END_OF_BACKUPSET)
	break;
	*/
      /* Try to read again.  If it was caused by EOT, there should be an EOT.
       * If not, (ie the next MtRead does NOT return MT_EOT), it must have been
       * a read error (?), so abort and bitch.
       */
      if (BurtMtRead(recoverMCB->inputChannel, &packet) == MT_EOT) 
      {
	if (LoadNextRecoverTape(recoverMCB->inputChannel, 
				recoverMCB->logChannel) == TCL_ERROR) 
	{
	  sprintf(logBuffer, "BURT: I/O error during recover, couldn't load "
		  "next tape [%s]\n", BurtTime());
	  Tcl_Write(recoverMCB->logChannel, logBuffer, strlen(logBuffer));
	  Tcl_Flush(recoverMCB->logChannel);
	  return;
	}
	continue;
      } 
      else 
      {
	sprintf(logBuffer, "BURT: Read error during recover, aborting [%s]\n",
		BurtTime());
	Tcl_Write(recoverMCB->logChannel, logBuffer, strlen(logBuffer));
	Tcl_Flush(recoverMCB->logChannel);
	return;
      }	    
    } 
    else if (nbytes > 0) 
    {
      if (packet.blockSize == END_OF_BACKUPSET) 
      {
	sprintf(logBuffer, "BURT: End of backupset in recover [%s]\n", 
		BurtTime());
	Tcl_Write(recoverMCB->logChannel, logBuffer, strlen(logBuffer));
	Tcl_Flush(recoverMCB->logChannel);
	return;
      }

      /* Set up the key for the hash table lookup */
      strncpy(key, packet.host, BURT_MAX_HOSTLEN);
      strcat(key, packet.atom);
      strcat(key, packet.level);
      strcat(key, packet.type);

      entryPtr = Tcl_FindHashEntry(&recoverCBTable, key);
      /* if entryPtr != NULL, then this is a packet for a scheduled atom */
      if (entryPtr != NULL)
      {
	recoverCB = (Burt_RecoverControlBlock *)Tcl_GetHashValue(entryPtr);

	if (recoverCB->outputChannel == NULL) 
	{
	  if (GetUserRecoverChannel(recoverMCB->interp, recoverCB) == TCL_ERROR)
	  {
	    /* Give up */
	    sprintf(logBuffer, "BURT: Unable to start %s %s %s %s, user defined "
		    "recover procedure returned TCL_ERROR [%s]\n",
		    recoverCB->host, recoverCB->atom, recoverCB->level,
		    recoverCB->type, BurtTime());
	    Tcl_Write(recoverMCB->logChannel, logBuffer, strlen(logBuffer));
	    Tcl_Flush(recoverMCB->logChannel);
	    Tcl_BackgroundError(recoverMCB->interp);
	    /* free memory allocated for the atom, remove it from hash table */
	    free((void *)recoverCB);
	    Tcl_DeleteHashEntry(entryPtr);
	    BurtStatusSetStarted(BurtStatusGetStarted() + 1);
	    BurtStatusSetFinished(BurtStatusGetFinished() + 1);
	  }
	  else
	  {
	    sprintf(logBuffer, "BURT: Started %s %s %s %s [%s]\n", 
		    recoverCB->host, recoverCB->atom, recoverCB->level, 
		    recoverCB->type, BurtTime());
	    Tcl_Write(recoverMCB->logChannel, logBuffer, strlen(logBuffer));
	    Tcl_Flush(recoverMCB->logChannel);
	    BurtStatusSetStarted(BurtStatusGetStarted() + 1);
	  }
	}
		
	if (packet.blockNum != ++(recoverCB->lastBlock)) 
	{
	  sprintf(logBuffer, "BURT: Aborting %s %s %s %s, incorrect block "
		  "sequence (got block %d, expected %d) [%s]\n",
		  recoverCB->host, recoverCB->atom, recoverCB->level, 
		  recoverCB->type, packet.blockNum, recoverCB->lastBlock, 
		  BurtTime());
	  Tcl_Write(recoverMCB->logChannel, logBuffer, strlen(logBuffer));
	  Tcl_Flush(recoverMCB->logChannel);

	  /* Close the atom's output channel */
	  Tcl_UnregisterChannel(recoverMCB->interp, recoverCB->outputChannel);
	  /* free memory allocated for the atom, remove it from the hash table */
	  free((void *)recoverCB);
	  Tcl_DeleteHashEntry(entryPtr);
	  BurtStatusSetFinished(BurtStatusGetFinished() + 1);
	}
	else
	{
	  /* 
	   * Check for end of dump, close up if we find it. 
	   */
	  
	  if (packet.blockSize == END_OF_BACKUP) 
	  {
	    sprintf(logBuffer, "BURT: Completed %s %s %s %s, %d blocks, "
		    "totaling %u bytes [%s]\n", packet.host, packet.atom, 
		    packet.level, packet.type, recoverCB->lastBlock, 
		    recoverCB->totalSize, BurtTime());
	    Tcl_Write(recoverMCB->logChannel, logBuffer, strlen(logBuffer));
	    Tcl_Flush(recoverMCB->logChannel);
	    Tcl_UnregisterChannel(recoverMCB->interp, recoverCB->outputChannel);
	    /* free memory allocated for the atom, remove it from hash table */
	    free((void *)recoverCB);
	    Tcl_DeleteHashEntry(entryPtr);
	    BurtStatusSetFinished(BurtStatusGetFinished() + 1);
	  } 
	  else if (packet.blockSize == ABORTED_BACKUP) 
	  {
	    sprintf(logBuffer, "BURT: Aborted %s %s %s %s [%s]\n",
		    packet.host, packet.atom, packet.level, packet.type,
		    BurtTime());
	    Tcl_Write(recoverMCB->logChannel, logBuffer, strlen(logBuffer));
	    Tcl_Flush(recoverMCB->logChannel);
	    Tcl_UnregisterChannel(recoverMCB->interp, recoverCB->outputChannel);
	    /* free memory allocated for the atom, remove it from hash table */
	    free((void *)recoverCB);
	    Tcl_DeleteHashEntry(entryPtr);
	    BurtStatusSetFinished(BurtStatusGetFinished() + 1);
	  } 
	  else if (packet.blockSize == KILLED_BACKUP) 
	  {
	    sprintf(logBuffer, "BURT: Killed %s %s %s %s [%s]\n",
		    packet.host, packet.atom, packet.level, packet.type,
		    BurtTime());
	    Tcl_Write(recoverMCB->logChannel, logBuffer, strlen(logBuffer));
	    Tcl_Flush(recoverMCB->logChannel);
	    Tcl_UnregisterChannel(recoverMCB->interp, recoverCB->outputChannel);
	    /* free memory allocated for the atom, remove it from hash table */
	    free((void *)recoverCB);
	    Tcl_DeleteHashEntry(entryPtr);
	    BurtStatusSetFinished(BurtStatusGetFinished() + 1);
	  } 
	  else 
	  {
	    if (packet.checksum != BurtChecksum((void *)packet.buffer,
						packet.blockSize))
	    {
	      sprintf(logBuffer, "BURT: Aborting %s %s %s %s, "
		      "incorrect checksum in block %d [%s]\n",
		      packet.host, packet.atom, packet.level, packet.type, 
		      packet.blockNum, BurtTime());
	      Tcl_Write(recoverMCB->logChannel, logBuffer, strlen(logBuffer));
	      Tcl_Flush(recoverMCB->logChannel);
	      Tcl_UnregisterChannel(recoverMCB->interp, 
				    recoverCB->outputChannel);
	      /* free memory allocated for the atom, remove it from hash table */
	      free((void *)recoverCB);
	      Tcl_DeleteHashEntry(entryPtr);
	      BurtStatusSetFinished(BurtStatusGetFinished() + 1);
	    }
	    else
	    {
	      if (recoverCB->outputChannel == NULL)
	      {
		sprintf(logBuffer, "BURT: %s %s %s %s output channel is NULL "
			"[%s]\n", packet.host, packet.atom, packet.level, 
			packet.type, BurtTime());
		Tcl_Write(recoverMCB->logChannel, logBuffer, strlen(logBuffer));
		Tcl_Flush(recoverMCB->logChannel);
	      }
	      else
	      {
		nbytes = Tcl_Write(recoverCB->outputChannel, packet.buffer, 
				   packet.blockSize);
		if (nbytes < 0) 
		{
		  if (Tcl_GetErrno() == EPIPE) 
		  {
		    sprintf(logBuffer, "BURT: EPIPE on write for %s %s %s "
			    "%s [%s]\n", packet.host, packet.atom, packet.level,
			    packet.type, BurtTime());
		    Tcl_Write(recoverMCB->logChannel, logBuffer, 
			      strlen(logBuffer));
		    Tcl_Flush(recoverMCB->logChannel);
		  
		    /*
		     * We don't consider a SIGPIPE an error due to
		     * the fact it happens all the time when /etc/restore
		     * finishes without reading all the data on tape.
		     */
		    Tcl_UnregisterChannel(recoverMCB->interp, 
					  recoverCB->outputChannel);
		    /* free memory for the atom, remove it from hash table */
		    free((void *)recoverCB);
		    Tcl_DeleteHashEntry(entryPtr);
		    BurtStatusSetFinished(BurtStatusGetFinished() + 1);
		  } 
		  else
		  { 
		    sprintf(logBuffer, "BURT: Aborting %s %s %s %s, write error "
			    "[%s]\n", recoverCB->host, recoverCB->atom, 
			    recoverCB->level, recoverCB->type, BurtTime());
		    Tcl_Write(recoverMCB->logChannel, logBuffer, -1);
		    Tcl_Flush(recoverMCB->logChannel);
		    Tcl_UnregisterChannel(recoverMCB->interp, 
					  recoverCB->outputChannel);
		    /* free memory for the atom, remove it from hash table */
		    free((void *)recoverCB);
		    Tcl_DeleteHashEntry(entryPtr);
		    BurtStatusSetFinished(BurtStatusGetFinished() + 1);
		  }
		}
		recoverCB->totalSize += nbytes;
		BurtStatusSetBytes(BurtStatusGetBytes() + nbytes);
	      }
	    }
	  }
	}
      }
    }
    if (BurtRecoverIsComplete())
    {
      BurtStatusSetRunState(RECOVER | COMPLETE);
      return;
    }
  } while (packet.blockSize != END_OF_BACKUPSET);
  return;
}

/*
 *----------------------------------------------------------------------
 * 
 * BurtRecoverIsComplete --
 * 
 *      This procedure checks to see if the recover is complete
 *
 * Results:
 *      1 if the recover session is complete, 0 otherwise.
 *
 * Side effects:
 *      None.
 *---------------------------------------------------------------------
 */

static int
BurtRecoverIsComplete()
{
  /* Unlike in backup, there's only one condition for completion:  
   * number of scheduled atoms == number of completed atoms
   */

  if (BurtStatusGetScheduled() == BurtStatusGetFinished())
    return 1;
  
  return 0;
}

/*
 *----------------------------------------------------------------------
 *
 * LoadNextRecoverTape --
 *
 *	This procedure prompts the user (via the logChannel) to mount the
 *      next tape, Before doing so it rewinds and unmounts the old one.  
 *      This procedure blocks until it gets a new tape and is able to read 
 *      a tape label. 
 *
 * Results:
 *	
 *      TCL_OK if the tape was swapped and the next tape label was
 *      read.
 *
 * Side effects:
 *      
 *----------------------------------------------------------------------
 */

int 
LoadNextRecoverTape(Tcl_Channel mtChannel, Tcl_Channel logChannel)
{
  int fd;               /* Unix level file descriptor. */
  char logBuffer[1024]; /* Holds pretty log output */
  int foundEvent;
  Burt_LoadNextTapeData promptData;

  if (BurtMtSafeRewind(mtChannel) == MT_ERROR) 
    return (TCL_ERROR);
  
  if (BurtMtSafeOffline(mtChannel) == MT_ERROR) 
    return (TCL_ERROR);
  
  /*
   * SunOS pukes at 2GB file size limit; reset seek pointer to 0
   */
  Tcl_GetChannelHandle(mtChannel, TCL_READABLE, (ClientData *) &fd);
  lseek(fd, 0, SEEK_SET);
  
  sprintf(logBuffer, "BURT: End of tape %u in recover [%s]\n", 
	  BurtStatusGetTapes(), BurtTime());
  Tcl_Write(logChannel, logBuffer, strlen(logBuffer));
  Tcl_Flush(logChannel);
  
  BurtStatusSetTapes(BurtStatusGetTapes() + 1);
  promptData.mtChannel = mtChannel;
  promptData.logChannel = logChannel;
  promptData.tapeReady = 0;
  
  sprintf(logBuffer, "BURT: Please mount tape %u (%d seconds) in recover... "
	  "[%s]\n", BurtStatusGetTapes(), TAPE_LOAD_DELAY, BurtTime());
  Tcl_Write(logChannel, logBuffer, -1);
  Tcl_Flush(logChannel);
  
  BurtStatusSetRunState(RECOVER | WAITING);
  promptData.promptTimer = Tcl_CreateTimerHandler(TAPE_LOAD_DELAY * 1000,
						  ProcessLoadRecoverTapeTimer,
						  (ClientData) &promptData);
  
  foundEvent = 1;
  while (!promptData.tapeReady && foundEvent)
    foundEvent = Tcl_DoOneEvent(0);

  Tcl_DeleteTimerHandler(promptData.promptTimer);
  
  BurtStatusSetRunState(RECOVER | RUN);
  sprintf(logBuffer, "BURT: Mounted tape %u in recover.  Thank you, "
	  "have a nice day [%s]\n", BurtStatusGetTapes(), BurtTime());
  Tcl_Write(logChannel, logBuffer, strlen(logBuffer));
  Tcl_Flush(logChannel);
  
  return(TCL_OK);
}

static void
ProcessLoadRecoverTapeTimer(ClientData clientData)
{
  Burt_LoadNextTapeData *data = (Burt_LoadNextTapeData *)clientData;
  Burt_TapeLabel tapeLabel;
  char logBuffer[1024];
  char tapeNo[3];
  char *label;

  if ((BurtMtRewind(data->mtChannel) == MT_ERROR) ||
      (BurtMtReadLabel(data->mtChannel, &tapeLabel) == MT_ERROR))
  {
    sprintf(logBuffer, "BURT: Please mount tape %u (%d seconds) in recover... "
	    "[%s]\n", BurtStatusGetTapes(), TAPE_LOAD_DELAY, BurtTime());
    Tcl_Write(data->logChannel, logBuffer, -1);
    Tcl_Flush(data->logChannel);
    data->promptTimer = Tcl_CreateTimerHandler(TAPE_LOAD_DELAY * 1000, 
					       ProcessLoadRecoverTapeTimer,
					       clientData);
  }
  else
  {
    /* Make sure they put the right tape in the drive! */
    sprintf(tapeNo, "%u", BurtStatusGetTapes());
    label = BurtStatusGetTapeLabel();
    if ((strncmp(tapeLabel.label, label, BURT_MAX_LABELLEN) != 0) || 
	(strncmp(tapeLabel.tape, tapeNo, 3) != 0))
    {
      sprintf(logBuffer, "BURT: Wrong tape in recover.  Expected tape %s "
	      "from session %s, got tape %s from session %s [%s]\n", 
	      tapeNo, label, tapeLabel.tape, tapeLabel.label, BurtTime());
      Tcl_Write(data->logChannel, logBuffer, -1);
      Tcl_Flush(data->logChannel);
      data->promptTimer = Tcl_CreateTimerHandler(TAPE_LOAD_DELAY * 1000, 
						 ProcessLoadRecoverTapeTimer,
						 clientData);
      BurtMtOffline(data->mtChannel);
    }
    else
    {
      data->tapeReady = 1;
    }
  }
  return;
}

Tcl_DString *
BurtRecoverRunningList(void)
{
  Tcl_DString *runningList;
  Tcl_HashEntry *entryPtr;
  Tcl_HashSearch search;
  Burt_RecoverControlBlock *recoverCB;
    
  runningList = (Tcl_DString *)malloc(sizeof(Tcl_DString));
  if (runningList != NULL) 
  {
    Tcl_DStringInit(runningList);
    entryPtr = Tcl_FirstHashEntry(&recoverCBTable, &search);
    while (entryPtr != NULL) 
    {
      recoverCB = (Burt_RecoverControlBlock *)Tcl_GetHashValue(entryPtr);
      if (recoverCB->outputChannel != NULL)
      {
	Tcl_DStringStartSublist(runningList);
	Tcl_DStringAppendElement(runningList, recoverCB->host);
	Tcl_DStringAppendElement(runningList, recoverCB->atom);
	Tcl_DStringAppendElement(runningList, recoverCB->level);
	Tcl_DStringAppendElement(runningList, recoverCB->type);
	Tcl_DStringEndSublist(runningList);
      }
      entryPtr = Tcl_NextHashEntry(&search);
    }
  }

  return(runningList);
}
