/*
 * burtSchedule.c
 *
 *    This file implements the schedule module.  It allows schedule items
 *    to be added, deleted, listed, etc...  This is where the Tcl command
 *    "schedule" is found.
 *
 * Copyright (C) 1995 - 1998 Eric Melski and Dean Jansa
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice and this permission notice appear in all copies,
 * and that all support documentation is included with each copy of the software.
 * A reasonable copying fee may be charged for any distribution of this 
 * software, but no fee may be charged for the software itself. You may 
 * distribute this software in aggregate with other programs as part of a 
 * larger software distribution provided that this software is not advertised 
 * as a product of your own.  Eric Melski and Dean Jansa make no representations
 * about the suitability of this software for any purpose.  It is provided 
 * "as is" without express or implied warranty.
 *
 * ERIC MELSKI AND DEAN JANSA DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS.
 * IN NO EVENT SHALL ERIC MELSKI OR DEAN JANSA BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 */

#define _POSIX_SOURCE
#ifndef lint
static char rcsid[] = "$Id: burtSchedule.c,v 1.9 1998/08/18 20:39:44 jmelski Exp $";
#endif

#include <burtInt.h>

Tcl_HashTable scheduleTable;   /* Table of scheduled dumps */


/*
 *----------------------------------------------------------------------
 *
 * Burt_ScheduleCmd --
 *
 *	This procedure implements the Tcl command schedule.
 *
 * Results:
 *	Returns a standard Tcl completion code, and leaves an error
 *	message in interp->result if an error occurs.
 *
 * Side effects:
 *       scheduleTable may be modified.
 *
 *----------------------------------------------------------------------
 */

int 
Burt_ScheduleCmd(ClientData cdata, Tcl_Interp *interp, 
			    int argc, char **argv)
{
    Burt_ScheduleItem *scheduleItemPtr; /* Item to add/remove to/from table. */
    Tcl_DString *scheduleList;
    char key[BURT_MAX_HOSTLEN + BURT_MAX_ATOMLEN + 
	    BURT_MAX_LEVELLEN + BURT_MAX_TYPELEN + 1];
    int length;
    char c;
    
    if (argc < 2) {
	Tcl_SetResult (interp, 
		       "wrong # args: should be \"schedule option ?arg ...?\"",
		       TCL_STATIC);
	return(TCL_ERROR);
    }
    
    c = argv[1][0];
    length = strlen(argv[1]); 
    if ((c == 'a') && (strncmp(argv[1], "add", length) == 0)) {
	if (argc != 6) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
			     " add host atom level type\"", (char *) NULL);
	    return(TCL_ERROR);
	}
	
	scheduleItemPtr = (Burt_ScheduleItem *)malloc(sizeof(Burt_ScheduleItem));
	/* 
         * Stuff args into an Burt_ScheduleItem struct and enqueue it.
         */
	strncpy(scheduleItemPtr->host, argv[2], BURT_MAX_HOSTLEN);
	strncpy(scheduleItemPtr->atom, argv[3], BURT_MAX_ATOMLEN);
	strncpy(scheduleItemPtr->level, argv[4], BURT_MAX_LEVELLEN);
	strncpy(scheduleItemPtr->type, argv[5], BURT_MAX_TYPELEN);
	
	strncpy(key, argv[2], BURT_MAX_HOSTLEN);
	strcat(key, argv[3]);
	strcat(key, argv[4]);
	strcat(key, argv[5]);

	BurtScheduleAdd(&scheduleTable, key, scheduleItemPtr);
	return TCL_OK;                                
    } else if ((c == 'r') && (strncmp(argv[1], "remove", length) == 0)) {
	if (argc != 6) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
			     " remove host atom level type\"", (char *) NULL);
	    return(TCL_ERROR);
	} else {
	    strncpy(key, argv[2], BURT_MAX_HOSTLEN);
	    strcat(key, argv[3]);
	    strcat(key, argv[4]);
	    strcat(key, argv[5]);
	    
	    if (BurtScheduleRemove(&scheduleTable, key) != TCL_OK) {
		Tcl_AppendResult(interp, "Could not find ", argv[2], 
				 " ", argv[3], " ", argv[4], " ",
				 argv[5], (char *) NULL);
		return(TCL_ERROR);
	    }

	    return(TCL_OK);
	}
    } else if ((c == 'l') && (strncmp(argv[1], "length", length) == 0)) {
	if (argc != 2) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
			     " length\"", (char *) NULL);
	    return(TCL_ERROR);
	} else {
	    sprintf(interp->result, "%d", BurtTableLength(&scheduleTable));
	    return TCL_OK;
	}
    } else if ((c == 'l') && (strncmp(argv[1], "list", length) == 0)) {
	if (argc != 2) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
			     " list\"", (char *) NULL);
	    return(TCL_ERROR);
	} else {
	    scheduleList = BurtScheduleList(&scheduleTable);
	    if (scheduleList == NULL) {
		Tcl_SetResult(interp, "Could not allocate memory for list!",
			      TCL_STATIC);
		return(TCL_ERROR);
	    } else {
		Tcl_DStringResult(interp, scheduleList);
		Tcl_DStringFree(scheduleList);
		return(TCL_OK);
	    }
	}
    } else if ((c == 'i') && (strncmp(argv[1], "info", length) == 0)) {
	if (argc != 2) {
	    Tcl_AppendResult(interp, "wrong # args: should be \"", argv[0],
			     " info\"", (char *) NULL);
	    return(TCL_ERROR);
	} else {
	    Tcl_SetResult(interp, BurtScheduleInfo(&scheduleTable), 
			  TCL_DYNAMIC);
	    return TCL_OK;
	}
    } else {
	Tcl_AppendResult(interp, "bad option \"", argv[1],
			 "\": should be add, remove, length, ",
			 "list or info", (char *) NULL);
	return TCL_ERROR;    
    }
}



/*
 *----------------------------------------------------------------------
 *
 * BurtScheduleInit --
 *
 *	This procedure sets up the schedule table.
 *
 * Results:
 *	None
 *
 * Side effects:
 *       *
 *----------------------------------------------------------------------
 */

void
BurtScheduleInit(void)
{
    Tcl_InitHashTable(&scheduleTable, TCL_STRING_KEYS);
    return;
}


/*
 *----------------------------------------------------------------------
 *
 * BurtScheduleAdd --
 *
 *	This procedure adds a Burt_ScheduleData struct to the
 *      schedule table.
 *
 * Results:
 *	Returns a standard Tcl completion code.
 *
 * Side effects:
 *       scheduleTable grows.
 *
 *----------------------------------------------------------------------
 */

int
BurtScheduleAdd (Tcl_HashTable *tablePtr, char *key, 
		 Burt_ScheduleItem *scheduleItemPtr)
{
    Tcl_HashEntry *entryPtr;
    int new;

    entryPtr = Tcl_CreateHashEntry(tablePtr, key, &new);

    if (new) {
	Tcl_SetHashValue(entryPtr, scheduleItemPtr);
    } else {
	/* 
	 * There is already an entry in the table for this
         * schedule item.
	 */
	return(TCL_ERROR);
    }

    return(TCL_OK);
}



/*
 *----------------------------------------------------------------------
 *
 * BurtScheduleRemove --
 *
 *	This procedure removes a Burt_ScheduleData struct from the
 *      schedule table. It does not return it, rather it throws it
 *      away.
 *
 * Results:
 *	Returns a standard Tcl completion code.
 *
 * Side effects:
 *       scheduleTable shrinks.
 *
 *----------------------------------------------------------------------
 */

int
BurtScheduleRemove(Tcl_HashTable *tablePtr, char *key)
{
    Tcl_HashEntry *entryPtr;
    
    entryPtr = Tcl_FindHashEntry(tablePtr, key);
    if (entryPtr != NULL) {
	Tcl_DeleteHashEntry(entryPtr);
	free((void *)entryPtr);
    } else {
	return(TCL_ERROR);
    }

    return(TCL_OK);
}



/*
 *----------------------------------------------------------------------
 *
 * BurtScheduleGetNext --
 *
 *	This procedure returns the next scheduled item from the table.
 *      It also removes thie entry in the schedule table.
 *
 * Results:
 *	Returns the item to caller, it is the callers responsibility to
 *      free the storage.  NULL is returned if the table is empty.
 *
 * Side effects:
 *      scheduleTable shrinks.
 *----------------------------------------------------------------------
 */

Burt_ScheduleItem *
BurtScheduleGetNext(Tcl_Interp *interp, Tcl_Channel logChannel, 
		    Tcl_HashTable *tablePtr)
{
    Burt_ScheduleItem *scheduleItemPtr;
    Tcl_HashEntry *entryPtr;
    Tcl_HashSearch search;

    char *orderingProc;
    int argc;
    char **argv;
    char key[BURT_MAX_HOSTLEN + BURT_MAX_ATOMLEN + 
	    BURT_MAX_LEVELLEN + BURT_MAX_TYPELEN + 1];


    orderingProc = Tcl_GetVar(interp, "burt_ordering_proc", TCL_GLOBAL_ONLY);
    if (orderingProc != NULL)
    {
      if (Tcl_VarEval(interp, orderingProc, NULL) == TCL_ERROR)
      {
	Tcl_Write(logChannel, "BURT: Error detected in user defined "
		  "ordering proc.", -1);
	entryPtr = Tcl_FirstHashEntry(tablePtr, &search);
      } 
      else
      {
	if (Tcl_SplitList(interp, interp->result, &argc, &argv) == TCL_ERROR)
	{
	  Tcl_Write(logChannel, 
		    "BURT: Error while parsing return from user "
		    "defined ordering proc.", -1);
	  entryPtr = Tcl_FirstHashEntry(tablePtr, &search);
	}
	else
	{
	  if (argc != 4)
	  {
	    if (argc > 0)
	    {
	      Tcl_Write(logChannel, 
			"BURT: Error in user defined ordering proc; "
			"must return a list of four elements.", -1);
	      Tcl_Flush(logChannel);
	    }
	    entryPtr = Tcl_FirstHashEntry(tablePtr, &search);
	  }
	  else 
	  {
	    strncpy(key, argv[0], BURT_MAX_HOSTLEN);
	    strcat(key, argv[1]);
	    strcat(key, argv[2]);
	    strcat(key, argv[3]);
	    entryPtr = Tcl_FindHashEntry(tablePtr, key);
	    Tcl_Free((char *)argv);
	    if (entryPtr == NULL) 
	    {
	      entryPtr = Tcl_FirstHashEntry(tablePtr, &search);
	    }
	  }
	}
      }
    }
    else
    {
      entryPtr = Tcl_FirstHashEntry(tablePtr, &search);
    }

    if (entryPtr == NULL) {
	return(NULL);
    }
    
    scheduleItemPtr = (Burt_ScheduleItem *)Tcl_GetHashValue(entryPtr);
    Tcl_DeleteHashEntry(entryPtr);
    return(scheduleItemPtr);
}



/*
 *----------------------------------------------------------------------
 *
 * BurtScheduleList --
 *
 *	This procedure creates a listing of all the scheduled items in
 *      the schedule table.
 *
 * Results:
 *	Returns a standard Tcl list of all the elements in the schedule table. 
 *      Null if the table is empty.  The caller is responsible for the 
 *      freeing of memory.  NOTE: It returns this list in a Tcl_DString *
 *      not just a char *!
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

Tcl_DString *
BurtScheduleList(Tcl_HashTable *tablePtr)
{
    Tcl_DString *scheduleList;
    Tcl_HashEntry *entryPtr;
    Tcl_HashSearch search;
    Burt_ScheduleItem *scheduleItemPtr;

    scheduleList = (Tcl_DString *)malloc(sizeof(Tcl_DString));
    if (scheduleList != NULL) {
	Tcl_DStringInit(scheduleList);
	entryPtr = Tcl_FirstHashEntry(tablePtr, &search);
	while (entryPtr != NULL) {
	    scheduleItemPtr = (Burt_ScheduleItem *)Tcl_GetHashValue(entryPtr);
	    Tcl_DStringStartSublist(scheduleList);
	    Tcl_DStringAppendElement(scheduleList, scheduleItemPtr->host);
	    Tcl_DStringAppendElement(scheduleList, scheduleItemPtr->atom);
	    Tcl_DStringAppendElement(scheduleList, scheduleItemPtr->level);
	    Tcl_DStringAppendElement(scheduleList, scheduleItemPtr->type);
	    Tcl_DStringEndSublist(scheduleList);
	    entryPtr = Tcl_NextHashEntry(&search);
	}
    }

    return(scheduleList);
}



/*
 *----------------------------------------------------------------------
 *
 * BurtScheduleInfo --
 *
 *	This procedure returns a dynamically-allocated string  with
 *      overall information about a hash table, such as the number
 *      of entries it contains, the number of buckets in its  hash
 *      array,  and  the  utilization  of  the buckets.
 *
 * Results:
 *     A dynamically-allocated string, it is the  caller's responsibility 
 *     to free the result string by passing it to free.    
 *
 * Side effects:
 *
 *----------------------------------------------------------------------
 */

char *
BurtScheduleInfo(Tcl_HashTable *tablePtr)
{
    return(Tcl_HashStats(tablePtr));
}

